from __future__ import annotations
from re import L
import pytest
import pytest_asyncio
import asyncio
import copy
from inspect import isawaitable
from moonraker.server import Server
from moonraker.utils import ServerError
from typing import TYPE_CHECKING, AsyncIterator, Dict, Any, Iterator

if TYPE_CHECKING:
    from components.database import MoonrakerDatabase
    from components.database import NamespaceWrapper
    from fixtures import HttpClient, WebsocketClient

TEST_DB: Dict[str, Dict[str, Any]] = {
    "automobiles": {
        "chevy": {
            "camaro": "silver",
            "silverado": {
                "1500": 3,
                "2500": 1
            }
        },
        "ford": {
            "mustang": "red",
            "f-series": {
                "f150": [150, "black"],
                "f350": {
                    "platinum": 10000,
                }
            }
        }
    },
    "fruits": {
        "apples": {
            "granny_smith": 10,
            "red_delicious": 8
        },
        "oranges": 50,
        "bananas": True
    },
    "vegetables": {
        "tomato": "nope"
    },
    "books": {
        "fantasy": {
            "lotr": "Gandalf"
        },
        "science_fiction": "dune"
    },
    "planets": {
        "earth": {
            "biosphere": True,
            "color": "blue"
        },
        "venus": {
            "hot": True
        },
        "mars": {
            "color": "red"
        },
        "jupiter": {
            "gas_giant": True,
            "europa": {
                "diameter": 3121.6
            },
            "io": "closest"
        },
        "saturn": {
            "has_rings": True
        },
        "pluto": "Don't unplanet me!"
    }
}

TEST_RECORD = {
    "debian": {
        "ubuntu": 10,
        "mint": True
    },
    "arch": 100,
    "redhat": {
        "centos": False
    }
}

TEST_OVERWRITE = {
    "vegetables": {
        "celery": "ranch",
        "lettuce": 100,
        "spinich": "popeye"
    },
    "oses": TEST_RECORD
}

@pytest_asyncio.fixture(scope="class")
async def base_db(base_server: Server) -> AsyncIterator[MoonrakerDatabase]:
    db: MoonrakerDatabase = base_server.load_component(
        base_server.config, "database")
    for ns, record in TEST_DB.items():
        for record_name, value in record.items():
            db.insert_item(ns, record_name, value)
    yield db
    await db.close()

@pytest_asyncio.fixture(scope="class")
async def running_db(base_server: Server) -> AsyncIterator[MoonrakerDatabase]:
    base_server.load_components()
    db: MoonrakerDatabase = base_server.lookup_component("database")
    for ns, record in TEST_DB.items():
        for record_name, value in record.items():
            db.insert_item(ns, record_name, value)
    await base_server.server_init(False)
    await base_server.start_server(False)
    yield db
    await base_server._stop_server("terminate")

# check_future() only resolves futures that are complete.  This
# is done to test database behavior in __init__() methods, where
# it is not possible to await a result.  We can't make this method
# async, as we need to check the future immediately.  Using an
# async would cause it to be scheduled on the event loop, with
# a thread potentially resolving a future before we can check it.
def check_future(fut: asyncio.Future,
                 db: MoonrakerDatabase
                 ) -> Any:
    server = db.server
    if server.is_running():
        if fut.done():
            pytest.fail("Future done while server running")
        return fut
    elif not fut.done():
        pytest.fail("Future not ready before server start")
    return fut.result()

@pytest.mark.asyncio
class BaseTest:
    @pytest.fixture(scope="class")
    def db(self, base_db):
        return base_db

@pytest.mark.asyncio
class ThreadedTest:
    @pytest.fixture(scope="class")
    def db(self, running_db):
        return running_db

class TestInstantiation:
    @pytest.fixture(scope="class")
    def db(self,
           base_server: Server,
           event_loop: asyncio.AbstractEventLoop
           ) -> Iterator[MoonrakerDatabase]:
        db: MoonrakerDatabase
        db = base_server.load_component(base_server.config, "database")
        yield db
        event_loop.run_until_complete(db.close())

    def test_initial_state(self, db: MoonrakerDatabase):
        mrdb = db.get_item("moonraker").result()
        assert (
            list(db.namespaces.keys()) == ["moonraker"] and
            mrdb == {
                "database_version": 1,
                "database": {
                    "unsafe_shutdowns": 1
                }
            }
        )

    def test_wrap_invalid_namespace(self, db: MoonrakerDatabase):
        expected = "Namespace 'invalid' not found"
        with pytest.raises(ServerError, match=expected):
            db.wrap_namespace("invalid")

    def test_insert_record_nonetype(self, db: MoonrakerDatabase):
        ret = db._insert_record("moonraker", "test_key", None)
        assert ret is False

    def test_encode_error(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError, match="Error encoding val"):
            db._encode_value(set(["invalid_value"]))

    def test_decode_error(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError, match="Error decoding value"):
            db._decode_value(b"invalid")

class TestCoreServerLoaded:
    @pytest.fixture(scope="class")
    def db(self,
           base_server: Server,
           event_loop: asyncio.AbstractEventLoop
           ) -> Iterator[MoonrakerDatabase]:
        base_server.load_components()
        db: MoonrakerDatabase
        db = base_server.lookup_component("database")
        yield db
        event_loop.run_until_complete(
            base_server._stop_server("terminate"))

    def test_core_state(self, db: MoonrakerDatabase):
        mrdb = db.get_item("moonraker").result()
        expected_ns = ["gcode_metadata", "moonraker"]
        assert (
            sorted(db.namespaces.keys()) == expected_ns and
            mrdb == {
                "database_version": 1,
                "database": {
                    "protected_namespaces": expected_ns,
                    "unsafe_shutdowns": 1
                },
                "file_manager": {
                    "metadata_version": 3
                }
            }
        )

@pytest.mark.run_paths(database="bare_db.cdb")
class TestCoreServerPreloaded(TestCoreServerLoaded):
    def test_core_state(self, db: MoonrakerDatabase):
        expected_ns = [
            "moonraker", "gcode_metadata", "update_manager",
            "authorized_users", "history"
        ]
        mrdb = db.get_item("moonraker").result()
        assert (
            sorted(db.namespaces.keys()) == sorted(expected_ns) and
            mrdb["database"]["unsafe_shutdowns"] == 2
        )

class TestUnallowedMethods:
    def test_register_error(self, running_db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            running_db.register_local_namespace("fruits")

    def test_wrap_namespace(self, running_db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            running_db.wrap_namespace("fruits")

class TestInsertItem(BaseTest):
    async def test_insert_record(self, db: MoonrakerDatabase):
        db.insert_item("oses", "linux", TEST_RECORD)
        fut = db.get_item("oses", "linux")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_RECORD

    async def test_insert_nested(self, db: MoonrakerDatabase):
        db.insert_item(
            "oses", "windows.eleven.feburary.2022", "ok")
        fut = db.get_item("oses", "windows.eleven.feburary.2022")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == "ok"

    async def test_insert_nested_invalid_assign(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.insert_item("oses", "linux.arch.february", "2022")
            await ret

    async def test_insert_nested_reduce_failure(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.insert_item("oses", "linux.arch.february.2022", True)
            await ret

    async def test_overwrite_record(self, db: MoonrakerDatabase,
                                    caplog: pytest.LogCaptureFixture):
        db.insert_item("oses", "ios", 10)
        db.insert_item("oses", ["ios", "15.3"], True)
        fut = db.get_item("oses", "ios")
        expected_log = (
            "Warning: Key ios contains a value of type"
            " <class 'int'>. Overwriting with an object."
        )
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert (
            result == {"15.3": True} and
            expected_log in caplog.messages
        )

class TestInsertItemThreaded(ThreadedTest, TestInsertItem):
    pass
class TestGetItem(BaseTest):
    async def test_get_record(self, db: MoonrakerDatabase):
        fut = db.get_item("automobiles", "chevy")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["automobiles"]["chevy"]

    async def test_get_namespace(self, db: MoonrakerDatabase):
        fut = db.get_item("fruits")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["fruits"]

    async def test_get_namespace_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.get_item("trains")
            await ret

    async def test_get_namespace_default(self, db: MoonrakerDatabase):
        fut = db.get_item("trains", default={})
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {}

    async def test_get_record_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.get_item("automobiles", "toyota")
            await ret

    async def test_get_record_default(self, db: MoonrakerDatabase):
        fut = db.get_item("automobiles", "toyota", {})
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {}

    async def test_get_key_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.get_item("automobiles", "chevy.equinox")
            await ret

    async def test_get_key_fail_default(self, db: MoonrakerDatabase):
        fut = db.get_item("automobiles", "chevy.equinox", "suv")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == "suv"

    async def test_get_nested(self, db: MoonrakerDatabase):
        fut = db.get_item(
            "automobiles", "ford.f-series.f350.platinum")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == 10000

    async def test_get_nested_no_key(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.get_item("automobiles", "ford.f-series.f350.superduty")
            await ret

    async def test_get_nested_no_key_default(self, db: MoonrakerDatabase):
        fut = db.get_item(
            "automobiles", "ford.f-series.f350.superduty", "success")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == "success"

    async def test_get_record_invalid_key(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.get_item("fruits", "apples..red_delicious")
            await ret

    async def test_get_record_invalid_key_type(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.get_item("vegetables", 100)
            await ret

    async def test_get_record_key_list(self, db: MoonrakerDatabase):
        key = ["chevy", "silverado", "2500"]
        fut = db.get_item("automobiles", key)
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == 1

class TestGetItemThreaded(ThreadedTest, TestGetItem):
    pass

class TestUpdateItem(BaseTest):
    async def test_update_record(self, db: MoonrakerDatabase):
        update_val = {
            "granny_smith": 1000,
            "jazz": 10.8,
            "gala": {"bland": True}
        }
        db.update_item("fruits", "apples", update_val)
        fut = db.get_item("fruits", "apples")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {
            "granny_smith": 1000,
            "red_delicious": 8,
            "jazz": 10.8,
            "gala": {"bland": True}
        }

    async def test_update_nested(self, db: MoonrakerDatabase):
        update_val = {"3500": {"color": "green"}, "2500": None}
        db.update_item("automobiles", "chevy.silverado", update_val)
        fut = db.get_item("automobiles", "chevy")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {
            "camaro": "silver",
            "silverado": {
                "1500": 3,
                "2500": None,
                "3500": {"color": "green"}
            }
        }

    async def test_update_replace_nested(self, db: MoonrakerDatabase):
        db.update_item("fruits", "apples.gala.bland", "ok")
        fut = db.get_item("fruits", "apples")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {
            "granny_smith": 1000,
            "red_delicious": 8,
            "jazz": 10.8,
            "gala": {"bland": "ok"}
        }

    async def test_update_replace_nested_dict(self, db: MoonrakerDatabase):
        db.update_item("automobiles", "ford.f-series.f350", "tow")
        fut = db.get_item("automobiles", "ford")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {
            "mustang": "red",
            "f-series": {
                "f150": [150, "black"],
                "f350": "tow"
            }
        }

    async def test_update_namespace_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_item("pizza", "deepdish", {})
            await ret

    async def test_update_record_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_item("automobiles", "toyota", {})
            await ret

    async def test_update_replace_record_dict_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_item("fruits", "apples", None)
            await ret

    async def test_update_replace_record_dict(self, db: MoonrakerDatabase):
        db.update_item("fruits", "apples", ["success"])
        fut = db.get_item("fruits", "apples")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == ["success"]

    async def test_update_key_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_item("automobiles", "ford.raptor", 10)
            await ret

    async def test_update_nested_key_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_item("automobiles", "ford.mustang.cobra", 10)
            await ret

    async def test_update_nested_key_not_found(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_item("automobiles", "chevy.corvette.z06", 10)
            await ret

class TestUpdateItemThreaded(ThreadedTest, TestUpdateItem):
    pass


class TestDeleteItem(BaseTest):
    async def test_delete_nested_item(self, db: MoonrakerDatabase):
        del_fut = db.delete_item(
            "automobiles", "ford.f-series.f350.platinum")
        del_result = check_future(del_fut, db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = db.get_item("automobiles", "ford")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert (
            del_result == 10000 and
            result == {
                "mustang": "red",
                "f-series": {
                    "f150": [150, "black"],
                    "f350": {}
                }
            }
        )

    async def test_delete_nested_dict(self, db: MoonrakerDatabase):
        del_fut = db.delete_item(
            "automobiles", "ford.f-series.f350")
        del_result = check_future(del_fut, db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = db.get_item("automobiles", "ford")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert (
            del_result == {} and
            result == {
                "mustang": "red",
                "f-series": {
                    "f150": [150, "black"],
                }
            }
        )

    async def test_delete_fail(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.delete_item("fruits", "bananas.green")
            await ret

    async def test_delete_record(self, db: MoonrakerDatabase):
        del_fut = db.delete_item("fruits", "bananas")
        del_result = check_future(del_fut, db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = db.get_item("fruits", "bananas", None)
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert (
            del_result is True and
            result is None
        )

    async def test_delete_last_nested(self, db: MoonrakerDatabase):
        del_fut = db.delete_item("books", "fantasy.lotr")
        del_result = check_future(del_fut, db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = db.get_item("books", "fantasy", None)
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert del_result == "Gandalf" and result is None

    async def test_drop_db(self, db: MoonrakerDatabase):
        del_fut = db.delete_item("vegetables", "tomato",
                                 drop_empty_db=True)
        del_result = check_future(del_fut, db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = db.get_item("vegetables", default=None)
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert del_result == "nope" and result is None

class TestDeleteItemThreaded(ThreadedTest, TestDeleteItem):
    pass

class TestInsertBatch(BaseTest):
    async def test_insert_batch(self, db: MoonrakerDatabase):
        db.insert_batch("batch_test", TEST_DB)
        fut = db.get_item("batch_test")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB

    async def test_insert_batch_overwrite(self, db: MoonrakerDatabase):
        expected = copy.deepcopy(TEST_DB)
        expected.update(TEST_OVERWRITE)
        db.insert_batch("batch_test", TEST_OVERWRITE)
        fut = db.get_item("batch_test")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

class TestInsertBatchThreaded(ThreadedTest, TestInsertBatch):
    pass

class TestGetBatch(BaseTest):
    async def test_get_batch(self, db: MoonrakerDatabase):
        keys = ["apples", "oranges", "bananas"]
        fut = db.get_batch("fruits", keys)
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["fruits"]

    async def test_get_batch_invalid_namespace(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.get_batch("invalid", ["no", "key"])
            await fut

    async def test_get_batch_invalid_keys(self, db: MoonrakerDatabase):
        fut = db.get_batch("automobiles", ["chevy", "toyota", "dodge"])
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {"chevy": TEST_DB["automobiles"]["chevy"]}

    async def test_get_batch_no_valid_keys(self, db: MoonrakerDatabase):
        fut = db.get_batch("automobiles", ["toyota", "dodge"])
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {}

class TestGetBatchThreaded(ThreadedTest, TestGetBatch):
    pass

class TestMoveBatch(BaseTest):
    async def test_move_batch(self, db: MoonrakerDatabase):
        source_keys = list(TEST_DB["fruits"].keys())
        dest_keys = [f"super_{key}" for key in source_keys]
        expected = {dk: TEST_DB["fruits"][sk] for dk, sk in
                    zip(dest_keys, source_keys)}
        db.move_batch("fruits", source_keys, dest_keys)
        fut = db.get_item("fruits")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_move_batch_invalid_namespace(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.move_batch(
                "invalid_ns", ["super_banana", "super_apple"],
                ["banana", "apple"])
            await fut

    async def test_move_batch_invalid_key(self, db: MoonrakerDatabase):
        source_keys = ["chevy", "toyota", "dodge"]
        dest_keys = ["chevrolet", "lexus", "chrysler"]
        expected = copy.deepcopy(TEST_DB["automobiles"])
        expected["chevrolet"] = expected.pop("chevy")
        db.move_batch("automobiles", source_keys, dest_keys)
        fut = db.get_item("automobiles")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_move_batch_no_valid_keys(self, db: MoonrakerDatabase):
        db.move_batch("vegetables", ["celery", "peas"],
                      ["no_celery", "no_peas"])
        fut = db.get_item("vegetables")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["vegetables"]

    async def test_move_batch_mismatch_key_length(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.move_batch("books", ["science_fiction"],
                                ["science_fiction", "fantasy"])
            await ret

class TestMoveBatchThreaded(ThreadedTest, TestMoveBatch):
    pass

class TestDeleteBatch(BaseTest):
    async def test_delete_batch(self, db: MoonrakerDatabase):
        del_keys = ["mars", "venus", "pluto"]
        expected = copy.deepcopy(TEST_DB["planets"])
        for k in del_keys:
            del expected[k]
        db.delete_batch("planets", del_keys)
        fut = db.get_item("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_delete_batch_all_keys(self, db: MoonrakerDatabase):
        del_keys = (TEST_DB["automobiles"].keys())
        db.delete_batch("automobiles", del_keys)
        fut = db.get_item("automobiles")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {}

    async def test_delete_batch_invalid_namespace(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.delete_batch("invalid", ["no", "key"])
            await ret

    async def test_delete_batch_invalid_key(self, db: MoonrakerDatabase):
        del_keys = ["science_fiction", "horror", "documentary"]
        db.delete_batch("books", del_keys)
        fut = db.get_item("books")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {"fantasy": {"lotr": "Gandalf"}}

    async def test_delete_batch_no_valid_keys(self, db: MoonrakerDatabase):
        del_keys = ["grapes", "peaches", "strawberries"]
        db.delete_batch("fruits", del_keys)
        fut = db.get_item("fruits")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["fruits"]

class TestDeleteBatchThreaded(ThreadedTest, TestDeleteBatch):
    pass

class TestUpdateNamespace(BaseTest):
    async def test_update_namespace(self, db: MoonrakerDatabase):
        update_val = {
            "venus": {"hot": True},
            "pluto": {"dwarf": True},
            "uranus": "klignons",
            "mercury": [1, 2, 3]
        }
        db.update_namespace("planets", update_val)
        fut = db.get_item("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        expected = copy.deepcopy(TEST_DB["planets"])
        expected.update(update_val)
        assert result == expected

    async def test_update_namespace_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            ret = db.update_namespace("invalid", {"hello": False})
            await ret

class TestUpdateNamespaceThreaded(ThreadedTest, TestUpdateNamespace):
    pass

class TestClearNamespace(BaseTest):
    async def test_clear_namespace(self, db: MoonrakerDatabase):
        db.clear_namespace("fruits")
        fut = db.get_item("fruits")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == {}

    async def test_clear_namespace_drop(self, db: MoonrakerDatabase):
        fut = db.clear_namespace("books", drop_empty_db=True)
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert "books" not in db.namespaces


    async def test_clear_namespace_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.clear_namespace("invalid")
            await fut

class TestClearNamespaceThreaded(ThreadedTest, TestClearNamespace):
    pass

class TestSyncNamespace(BaseTest):
    async def test_sync_namespace(self, db: MoonrakerDatabase):
        synced = copy.deepcopy(TEST_DB["planets"])
        del synced["mars"]
        del synced["pluto"]
        synced.update({"mercury": "close", "neptune": "far",
                      "venus": "cloudy"})
        db.sync_namespace("planets", synced)
        fut = db.get_item("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == synced

    async def test_sync_no_overlap(self, db: MoonrakerDatabase):
        synced = {"toyota": {"corolla": "car", "tundra": "truck"}}
        db.sync_namespace("automobiles", synced)
        fut = db.get_item("automobiles")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == synced

    async def test_sync_no_remove(self, db: MoonrakerDatabase):
        synced = copy.deepcopy(TEST_DB["fruits"])
        synced.update({"cherries": "sweet", "berries": {"blue": "mild"}})
        db.sync_namespace("fruits", synced)
        fut = db.get_item("fruits")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == synced

    async def test_sync_namespace_empty(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.sync_namespace("books", {})
            await fut

    async def test_sync_namespace_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.sync_namespace("invalid", {"no": "key"})
            await fut

class TestSyncNamespaceThreaded(ThreadedTest, TestSyncNamespace):
    pass

class TestNamespaceLength(BaseTest):
    async def test_ns_length(self, db: MoonrakerDatabase):
        expected = len(TEST_DB["planets"])
        fut = db.ns_length("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_ns_length_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.ns_length("invalid")
            await fut

class TestNamespaceLengthThreaded(ThreadedTest, TestNamespaceLength):
    pass

class TestNamespaceKeys(BaseTest):
    async def test_ns_keys(self, db: MoonrakerDatabase):
        expected = list(TEST_DB["planets"].keys())
        fut = db.ns_keys("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == sorted(expected)

    async def test_ns_keys_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.ns_keys("invalid")
            await fut

class TestNamespaceKeysThreaded(ThreadedTest, TestNamespaceKeys):
    pass

class TestNamespaceValues(BaseTest):
    async def test_ns_values(self, db: MoonrakerDatabase):
        expected = [i[1] for i in sorted(TEST_DB["planets"].items(),
                    key=lambda d: d[0])]
        fut = db.ns_values("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_ns_values_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.ns_values("invalid")
            await fut

class TestNamespaceValuesThreaded(ThreadedTest, TestNamespaceValues):
    pass

class TestNamespaceItems(BaseTest):
    async def test_ns_items(self, db: MoonrakerDatabase):
        expected = sorted(TEST_DB["planets"].items(), key=lambda d: d[0])
        fut = db.ns_items("planets")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_ns_items_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.ns_items("invalid")
            await fut

class TestNamespaceItemsThreaded(ThreadedTest, TestNamespaceItems):
    pass

class TestNamespaceContains(BaseTest):
    async def test_ns_contains_record(self, db: MoonrakerDatabase):
        fut = db.ns_contains("planets", "venus")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result is True

    async def test_ns_not_contains_record(self, db: MoonrakerDatabase):
        fut = db.ns_contains("planets", "mercury")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result is False

    async def test_ns_contains_nested(self, db: MoonrakerDatabase):
        fut = db.ns_contains("automobiles", "ford.f-series.f350.platinum")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result is True

    async def test_ns_not_contains_nested(self, db: MoonrakerDatabase):
        fut = db.ns_contains("automobiles", "ford.f-series.f250.fx4")
        result = check_future(fut, db)
        if isawaitable(result):
            result = await result
        assert result is False

    async def test_ns_contains_invalid(self, db: MoonrakerDatabase):
        with pytest.raises(ServerError):
            fut = db.ns_contains("invalid", "nokey")
            await fut

class TestNamespaceConainsThreaded(ThreadedTest, TestNamespaceContains):
    pass

class WrapperTest(BaseTest):
    @pytest.fixture(scope="class")
    def wrapped(self,
                request: pytest.FixtureRequest,
                db: MoonrakerDatabase
                ) -> NamespaceWrapper:
        parse = not request.cls.__name__.endswith("NoParse")
        return db.wrap_namespace("planets", parse_keys=parse)

@pytest.mark.asyncio
class WrapperTestThreaded:
    @pytest_asyncio.fixture(scope="class")
    async def wrapped(self,
                      request: pytest.FixtureRequest,
                      base_server: Server
                      ) -> AsyncIterator[MoonrakerDatabase]:
        base_server.load_components()
        db: MoonrakerDatabase = base_server.lookup_component("database")
        for ns, record in TEST_DB.items():
            for record_name, value in record.items():
                db.insert_item(ns, record_name, value)
        parse = not request.cls.__name__.endswith("NoParse")
        wrapped = db.wrap_namespace("planets", parse_keys=parse)
        await base_server.server_init(False)
        await base_server.start_server(False)
        yield wrapped
        await base_server._stop_server("terminate")

    async def test_asdict(self, wrapped: NamespaceWrapper):
        with pytest.raises(ServerError):
            wrapped.as_dict()

    async def test_contains_magic(self, wrapped: NamespaceWrapper):
        with pytest.raises(ServerError):
            "earth" in wrapped


class TestNamespaceWrapper(WrapperTest):
    async def test_wrapped_insert(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        expected["oses"] = TEST_RECORD
        wrapped.insert("oses", TEST_RECORD)
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_wrapped_get(self, wrapped: NamespaceWrapper):
        fut = wrapped.get("oses")
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == TEST_RECORD

    async def test_wrapped_delete(self, wrapped: NamespaceWrapper):
        del_fut = wrapped.delete("oses")
        del_result = check_future(del_fut, wrapped.db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert (
            del_result == TEST_RECORD and
            result == TEST_DB["planets"]
        )

    async def test_wrapped_nested_insert(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        if wrapped.parse_keys:
            expected["oses"] = {"nested": TEST_RECORD}
        else:
            expected["oses.nested"] = TEST_RECORD
        wrapped.insert("oses.nested", TEST_RECORD)
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_nested_get(self, wrapped: NamespaceWrapper):
        fut = wrapped.get("oses.nested")
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == TEST_RECORD

    async def test_wrapped_nested_delete(self, wrapped: NamespaceWrapper):
        del_fut = wrapped.delete("oses.nested")
        del_result = check_future(del_fut, wrapped.db)
        if isawaitable(del_result):
            del_result = await del_result
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert (
            del_result == TEST_RECORD and
            result == TEST_DB["planets"]
        )

    async def test_update_child(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        expected["pluto"] = {"type": "dwarf"}
        wrapped.update_child("pluto", {"type": "dwarf"})
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_update_child_nested(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        expected["pluto"] = {"type": "planet!"}
        wrapped.update_child("pluto.type", "planet!")
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_update(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        upval = {"pluto": "Don't unplanet me!", "caprica": {"bsg": True}}
        expected.update(upval)
        wrapped.update(upval)
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_sync(self, wrapped: NamespaceWrapper):
        wrapped.sync(TEST_DB["planets"])
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["planets"]

    async def test_insert_batch(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        expected.update(TEST_RECORD)
        wrapped.insert_batch(TEST_RECORD)
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_get_batch(self, wrapped: NamespaceWrapper):
        fut = wrapped.get_batch(list(TEST_RECORD.keys()))
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == TEST_RECORD

    async def test_move_batch(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        expected.update({f"{k}_moved": v for k, v in TEST_RECORD.items()})
        source_keys = list(TEST_RECORD.keys())
        dest_keys = [f"{key}_moved" for key in source_keys]
        wrapped.move_batch(source_keys, dest_keys)
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_delete_batch(self, wrapped: NamespaceWrapper):
        expected = {f"{k}_moved": v for k, v in TEST_RECORD.items()}
        fut = wrapped.delete_batch(list(expected.keys()))
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_length(self, wrapped: NamespaceWrapper):
        fut = wrapped.length()
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == len(TEST_DB["planets"])

    async def test_asdict(self, wrapped: NamespaceWrapper):
        assert wrapped.as_dict() == TEST_DB["planets"]

    async def test_setitem_magic(self, wrapped: NamespaceWrapper):
        expected = copy.deepcopy(TEST_DB["planets"])
        expected["neptune"] = {"distance_from_sun": 2.8}
        wrapped["neptune"] = {"distance_from_sun": 2.8}
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_getitem_magic(self, wrapped: NamespaceWrapper):
        fut = wrapped["neptune"]
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == {"distance_from_sun": 2.8}

    async def test_getitem_magic_error(self, wrapped: NamespaceWrapper):
        with pytest.raises(ServerError):
            fut = wrapped["orthos"]
            await fut

    async def test_del_magic(self, wrapped: NamespaceWrapper):
        del wrapped["neptune"]
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["planets"]

    async def test_contains_magic(self, wrapped: NamespaceWrapper):
        assert "earth" in wrapped

    async def test_contains(self, wrapped: NamespaceWrapper):
        fut = wrapped.contains("earth")
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result is True

    async def test_contains_nested(self, wrapped: NamespaceWrapper):
        fut = wrapped.contains("jupiter.io")
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result is wrapped.parse_keys

    async def test_keys_method(self, wrapped: NamespaceWrapper):
        fut = wrapped.keys()
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == sorted(TEST_DB["planets"].keys())

    async def test_values_method(self, wrapped: NamespaceWrapper):
        expected = [TEST_DB["planets"][key] for key in
                    sorted(TEST_DB["planets"].keys())]
        fut = wrapped.values()
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_items_method(self, wrapped: NamespaceWrapper):
        expected = sorted(TEST_DB["planets"].items(), key=lambda x: x[0])
        fut = wrapped.items()
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == expected

    async def test_pop(self, wrapped: NamespaceWrapper):
        fut = wrapped.pop("jupiter")
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == TEST_DB["planets"]["jupiter"]

    async def test_pop_error(self, wrapped: NamespaceWrapper):
        with pytest.raises(ServerError):
            fut = wrapped.pop("invalid_key")
            await fut

    async def test_pop_default(self, wrapped: NamespaceWrapper):
        fut = wrapped.pop("invalid_key", "default_value")
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == "default_value"

    async def test_clear(self, wrapped: NamespaceWrapper):
        wrapped.clear()
        fut = wrapped.db.get_item(wrapped.namespace)
        result = check_future(fut, wrapped.db)
        if isawaitable(result):
            result = await result
        assert result == {}

class TestNamespaceWrapperNoParse(TestNamespaceWrapper):
    async def test_update_child_nested(self, wrapped: NamespaceWrapper):
        with pytest.raises(ServerError):
            ret = wrapped.update_child("pluto.type", "planet")
            await ret

class TestNamespaceWrapperThreaded(WrapperTestThreaded, TestNamespaceWrapper):
    pass

class TestNamespaceWrapperThreadedNoParse(
    WrapperTestThreaded, TestNamespaceWrapperNoParse
):
    pass

def endpoint_result(req_args: Dict[str, Any], expected: Any) -> Dict[str, Any]:
    return {
        "namespace": req_args["namespace"],
        "key": req_args.get("key"),
        "value": expected
    }

@pytest.mark.asyncio
class EndpointTest:
    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def server(self, base_server: Server) -> AsyncIterator[Server]:
        base_server.load_components()
        db: MoonrakerDatabase = base_server.lookup_component("database")
        for ns, record in TEST_DB.items():
            for record_name, value in record.items():
                db.insert_item(ns, record_name, value)
        db.register_local_namespace("planets")
        db.register_local_namespace("fruits", forbidden=True)
        await base_server.server_init(False)
        await base_server.start_server(False)
        yield base_server
        await base_server._stop_server("terminate")

    @pytest.fixture(scope="class")
    def db(self, server: Server) -> MoonrakerDatabase:
        return server.lookup_component("database")

class TestHttpEndpoints(EndpointTest):
    async def test_list_dbs(self, http_client: HttpClient):
        expected = list(TEST_DB.keys())
        expected.remove("fruits")
        expected.extend(["moonraker", "gcode_metadata"])
        ret = await http_client.get("/server/database/list")
        assert sorted(ret["result"]["namespaces"]) == sorted(expected)

    async def test_get_namespace(self, http_client: HttpClient):
        args = {"namespace": "automobiles"}
        ret = await http_client.get("/server/database/item", args)
        assert ret["result"] == endpoint_result(args, TEST_DB["automobiles"])

    async def test_get_namespace_not_exist(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 404:"):
            args = {"namespace": "cities"}
            await http_client.get("/server/database/item", args)

    async def test_get_item(self, http_client: HttpClient):
        args = {"namespace": "automobiles", "key": "ford.mustang"}
        ret = await http_client.get("/server/database/item", args)
        assert ret["result"] == endpoint_result(args, "red")

    async def test_get_item_not_exist(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 404:"):
            args = {"namespace": "automobiles", "key": "ford.mustang.year"}
            await http_client.get("/server/database/item", args)

    async def test_get_item_protected_ns(self, http_client: HttpClient):
        args = {"namespace": "planets", "key": "jupiter.gas_giant"}
        ret = await http_client.get("/server/database/item", args)
        assert ret["result"] == endpoint_result(args, True)

    async def test_get_item_forbidden_ns(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 403:"):
            args = {"namespace": "fruits", "key": "apples"}
            await http_client.get("/server/database/item", args)

    async def test_post_item(self, http_client: HttpClient,
                             db: MoonrakerDatabase):
        args = {"namespace": "breakfast", "key": "cereal.sweet",
                "value": "Count Chocula"}
        ret = await http_client.post("/server/database/item", args)
        check = await db.get_item("breakfast", "cereal.sweet")
        assert ret["result"] == args and check == "Count Chocula"

    async def test_post_item_no_key(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 400:"):
            args = {"namespace": "breakfast", "value": "pancakes"}
            await http_client.post("/server/database/item", args)

    async def test_post_item_no_value(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 400:"):
            args = {"namespace": "breakfast", "key": "pancakes"}
            await http_client.post("/server/database/item", args)

    async def test_post_item_protected(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 403:"):
            args = {"namespace": "planets", "key": "jupiter.gas_giant",
                    "value": "biggest"}
            await http_client.post("/server/database/item", args)

    async def test_post_item_forbidden(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 403:"):
            args = {"namespace": "fruits", "key": "cherries.color",
                    "value": "red"}
            await http_client.post("/server/database/item", args)

    async def test_delete_item(self, http_client: HttpClient,
                               db: MoonrakerDatabase):
        args = {"namespace": "automobiles", "key": "ford.f-series.f150"}
        ret = await http_client.delete("/server/database/item", args)
        check = await db.get_item("automobiles", "ford.f-series")
        assert (
            ret["result"] == endpoint_result(args, [150, "black"])
            and check == {"f350": {"platinum": 10000}}
        )

    async def test_delete_item_drop(self, http_client: HttpClient,
                                    db: MoonrakerDatabase):
        args = {"namespace": "vegetables", "key": "tomato"}
        ret = await http_client.delete("/server/database/item", args)
        assert (
            ret["result"] == endpoint_result(args, "nope")
            and "vegetables" not in db.namespaces
        )

    async def test_delete_item_not_found(self, http_client: HttpClient):
        with pytest.raises(http_client.error, match="HTTP 404: Not Found"):
            args = {"namespace": "automobiles", "key": "ford.pinto"}
            await http_client.delete("/server/database/item", args)

class TestWebsocketEndpoints(EndpointTest):
    async def test_list_dbs(self, websocket_client: WebsocketClient):
        expected = list(TEST_DB.keys())
        expected.remove("fruits")
        expected.extend(["moonraker", "gcode_metadata"])
        ret = await websocket_client.request("server.database.list")
        assert sorted(ret["namespaces"]) == sorted(expected)

    async def test_get_namespace(self, websocket_client: WebsocketClient):
        args = {"namespace": "automobiles"}
        ret = await websocket_client.request("server.database.get_item", args)
        assert ret == endpoint_result(args, TEST_DB["automobiles"])

    async def test_get_namespace_not_exist(self,
                                           websocket_client: WebsocketClient):
        expected = "Namespace 'cities' not found"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "cities"}
            await websocket_client.request("server.database.get_item", args)

    async def test_get_item(self, websocket_client: WebsocketClient):
        args = {"namespace": "automobiles", "key": "ford.mustang"}
        ret = await websocket_client.request("server.database.get_item", args)
        assert ret == endpoint_result(args, "red")

    async def test_get_item_not_exist(self,
                                      websocket_client: WebsocketClient):
        expected = (
            "Key 'ford.mustang.year' in namespace 'automobiles' not found"
        )
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "automobiles", "key": "ford.mustang.year"}
            await websocket_client.request("server.database.get_item", args)

    async def test_get_item_protected_ns(self,
                                         websocket_client: WebsocketClient):
        args = {"namespace": "planets", "key": "jupiter.gas_giant"}
        ret = await websocket_client.request("server.database.get_item", args)
        assert ret == endpoint_result(args, True)

    async def test_get_item_forbidden_ns(self,
                                         websocket_client: WebsocketClient):
        expected = "Read/Write access to namespace 'fruits' is forbidden"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "fruits", "key": "apples"}
            await websocket_client.request("server.database.get_item", args)

    async def test_post_item(self, websocket_client: WebsocketClient,
                             db: MoonrakerDatabase):
        args = {"namespace": "breakfast", "key": "cereal.sweet",
                "value": "Count Chocula"}
        ret = await websocket_client.request("server.database.post_item", args)
        check = await db.get_item("breakfast", "cereal.sweet")
        assert ret == args and check == "Count Chocula"

    async def test_post_item_no_key(self, websocket_client: WebsocketClient):
        expected = "No data for argument: key"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "breakfast", "value": "pancakes"}
            await websocket_client.request("server.database.post_item", args)

    async def test_post_item_no_value(self, websocket_client: WebsocketClient):
        expected = "No data for argument: value"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "breakfast", "key": "pancakes"}
            await websocket_client.request("server.database.post_item", args)

    async def test_post_item_protected(self,
                                       websocket_client: WebsocketClient):
        expected = "Write access to namespace 'planets' is forbidden"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "planets", "key": "jupiter.gas_giant",
                    "value": "biggest"}
            await websocket_client.request("server.database.post_item", args)

    async def test_post_item_forbidden(self,
                                       websocket_client: WebsocketClient):
        expected = "Read/Write access to namespace 'fruits' is forbidden"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "fruits", "key": "cherries.color",
                    "value": "red"}
            await websocket_client.request("server.database.post_item", args)

    async def test_delete_item(self, websocket_client: WebsocketClient,
                               db: MoonrakerDatabase):
        args = {"namespace": "automobiles", "key": "ford.f-series.f150"}
        ret = await websocket_client.request(
            "server.database.delete_item", args)
        check = await db.get_item("automobiles", "ford.f-series")
        assert (
            ret == endpoint_result(args, [150, "black"])
            and check == {"f350": {"platinum": 10000}}
        )

    async def test_delete_item_drop(self, websocket_client: WebsocketClient,
                                    db: MoonrakerDatabase):
        args = {"namespace": "vegetables", "key": "tomato"}
        ret = await websocket_client.request(
            "server.database.delete_item", args)
        assert (
            ret == endpoint_result(args, "nope")
            and "vegetables" not in db.namespaces
        )

    async def test_delete_item_not_found(self,
                                         websocket_client: WebsocketClient):
        expected = "Key 'ford.pinto' in namespace 'automobiles' not found"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "automobiles", "key": "ford.pinto"}
            await websocket_client.request("server.database.delete_item", args)

    async def test_invalid_key(self, websocket_client: WebsocketClient):
        expected = "Value for argument 'key' is an invalid type"
        with pytest.raises(websocket_client.error, match=expected):
            args = {"namespace": "planets", "key": {"ford": "pinto"}}
            await websocket_client.request("server.database.get_item", args)



#! /usr/bin/python3
# Script for syncing package dependencies and python reqs
#
# Copyright (C) 2024 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import argparse
import pathlib
import tomllib
import json
import re
from typing import Dict, List

MAX_LINE_LENGTH = 88
SCRIPTS_PATH = pathlib.Path(__file__).parent
INST_PKG_HEADER = "# *** AUTO GENERATED OS PACKAGE DEPENDENCES START ***"
INST_PKG_FOOTER = "# *** AUTO GENERATED OS PACKAGE DEPENDENCES END ***"

def gen_multline_var(var_name: str, values: List[str], indent: int = 0) -> str:
    idt = " " * indent
    if not values:
        return f'{idt}{var_name}=""'
    line_list: List[str] = []
    current_line = f"{idt}{var_name}=\"{values.pop(0)}"
    for val in values:
        if len(current_line) + len(val) + 2 > MAX_LINE_LENGTH:
            line_list.append(f'{current_line}"')
            current_line = (f"{idt}{var_name}=\"${{{var_name}}} {val}")
        else:
            current_line += f" {val}"
    line_list.append(f'{current_line}"')
    return "\n".join(line_list)

def sync_packages() -> int:
    inst_script = SCRIPTS_PATH.joinpath("install-moonraker.sh")
    sys_deps_file = SCRIPTS_PATH.joinpath("system-dependencies.json")
    new_deps: Dict[str, List[str]] = json.loads(sys_deps_file.read_bytes())
    # Copy install script in memory.
    install_data: List[str] = []
    prev_deps: Dict[str, List[str]] = {}
    distro_name = ""
    skip_data = False
    with inst_script.open("r") as inst_file:
        for line in inst_file:
            cur_line = line.strip()
            if not skip_data:
                install_data.append(line)
            else:
                # parse current dependencies
                distro_match = re.match(
                    r"(?:el)?if \[ \$\{DISTRIBUTION\} = \"([a-z0-9._-]+)\" \]; then",
                    cur_line
                )
                if distro_match is not None:
                    distro_name = distro_match.group(1)
                    prev_deps[distro_name] = []
                elif cur_line.startswith("PACKAGES"):
                    pkgs = cur_line.split("=", maxsplit=1)[1].strip('"')
                    pkg_list = pkgs.split()
                    if pkg_list and pkg_list[0] == "${PACKAGES}":
                        pkg_list.pop(0)
                    prev_deps[distro_name].extend(pkg_list)
            if cur_line == INST_PKG_HEADER:
                skip_data = True
            elif cur_line == INST_PKG_FOOTER:
                skip_data = False
                install_data.append(line)
    # Check if an update is necessary
    if set(prev_deps.keys()) == set(new_deps.keys()):
        for distro, pkg_list in prev_deps.items():
            new_pkgs = new_deps[distro]
            if set(pkg_list) != set(new_pkgs):
                break
        else:
            # Dependencies match, exit
            print("System package dependencies match")
            return 0
    print("Writing new system dependencies to install script...")
    with inst_script.open("w+") as inst_file:
        # Find and replace old package defs
        for line in install_data:
            inst_file.write(line)
            if line.strip() == INST_PKG_HEADER:
                indent_count = len(line) - len(line.lstrip())
                idt = " " * indent_count
                # Write Package data
                first_distro = True
                for distro, packages in new_deps.items():
                    prefix = f"{idt}if" if first_distro else f"{idt}elif"
                    first_distro = False
                    inst_file.write(
                        f'{prefix} [ ${{DISTRIBUTION}} = "{distro}" ]; then\n'
                    )
                    pkg_var = gen_multline_var("PACKAGES", packages, indent_count + 4)
                    inst_file.write(pkg_var)
                    inst_file.write("\n")
                inst_file.write(f"{idt}fi\n")
    return 1

def check_reqs_changed(reqs_file: pathlib.Path, new_reqs: List[str]) -> bool:
    req_list = []
    for requirement in reqs_file.read_text().splitlines():
        requirement = requirement.strip()
        if not requirement or requirement[0] in ("-", "#"):
            continue
        req_list.append(requirement)
    return set(new_reqs) != set(req_list)

def sync_requirements() -> int:
    ret: int = 0
    src_path = SCRIPTS_PATH.parent
    proj_file = src_path.joinpath("pyproject.toml")
    with proj_file.open("rb") as f:
        data = tomllib.load(f)
    python_deps = data["project"]["dependencies"]
    optional_deps = data["project"]["optional-dependencies"]
    reqs_path = SCRIPTS_PATH.joinpath("moonraker-requirements.txt")
    if check_reqs_changed(reqs_path, python_deps):
        print("Syncing Moonraker Python Requirements...")
        ret = 1
        with reqs_path.open("w+") as req_file:
            req_file.write("# Python dependencies for Moonraker\n")
            req_file.write("--find-links=python_wheels\n")
            for requirement in python_deps:
                req_file.write(f"{requirement}\n")
    else:
        print("Moonraker Python requirements match")
    # sync speedups
    speedups_path = SCRIPTS_PATH.joinpath("moonraker-speedups.txt")
    speedup_deps = optional_deps["speedups"]
    if check_reqs_changed(speedups_path, speedup_deps):
        print("Syncing speedup requirements...")
        ret = 1
        with speedups_path.open("w+") as req_file:
            for requirement in speedup_deps:
                req_file.write(f"{requirement}\n")
    else:
        print("Speedup sequirements match")
    # sync dev dependencies
    dev_reqs_path = SCRIPTS_PATH.joinpath("moonraker-dev-reqs.txt")
    dev_deps = optional_deps["dev"]
    if check_reqs_changed(dev_reqs_path, dev_deps):
        print("Syncing dev requirements")
        ret = 1
        with dev_reqs_path.open("r+") as req_file:
            for requirement in dev_deps:
                req_file.write(f"{requirement}\n")
    else:
        print("Dev requirements match")
    return ret

def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename", default="", nargs="?",
        help="The name of the dependency file to sync"
    )
    args = parser.parse_args()
    fname: str = args.filename
    if not fname:
        ret = sync_requirements()
        ret += sync_packages()
        return 1 if ret > 0 else 0
    elif fname == "pyproject.toml":
        return sync_requirements()
    elif fname == "scripts/system-dependencies.json":
        return sync_packages()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
# GCode metadata extraction utility
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import json
import argparse
import re
import os
import sys
import base64
import traceback
import tempfile
import zipfile
import shutil
import uuid
import logging
from PIL import Image

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Dict,
    List,
    Tuple,
    Type,
)
if TYPE_CHECKING:
    pass

UFP_MODEL_PATH = "/3D/model.gcode"
UFP_THUMB_PATH = "/Metadata/thumbnail.png"

logging.basicConfig(stream=sys.stderr, level=logging.INFO)
logger = logging.getLogger("metadata")

# Regex helpers.  These methods take patterns with placeholders
# to insert the correct regex capture group for floats, ints,
# and strings:
#  Float: (%F) = (\d*\.?\d+)
#  Integer: (%D) = (\d+)
#  String: (%S) = (.+)
def regex_find_floats(pattern: str, data: str) -> List[float]:
    pattern = pattern.replace(r"(%F)", r"([0-9]*\.?[0-9]+)")
    matches = re.findall(pattern, data)
    if matches:
        # return the maximum height value found
        try:
            return [float(h) for h in matches]
        except Exception:
            pass
    return []

def regex_find_ints(pattern: str, data: str) -> List[int]:
    pattern = pattern.replace(r"(%D)", r"([0-9]+)")
    matches = re.findall(pattern, data)
    if matches:
        # return the maximum height value found
        try:
            return [int(h) for h in matches]
        except Exception:
            pass
    return []

def regex_find_float(pattern: str, data: str) -> Optional[float]:
    pattern = pattern.replace(r"(%F)", r"([0-9]*\.?[0-9]+)")
    match = re.search(pattern, data)
    val: Optional[float] = None
    if match:
        try:
            val = float(match.group(1))
        except Exception:
            return None
    return val

def regex_find_int(pattern: str, data: str) -> Optional[int]:
    pattern = pattern.replace(r"(%D)", r"([0-9]+)")
    match = re.search(pattern, data)
    val: Optional[int] = None
    if match:
        try:
            val = int(match.group(1))
        except Exception:
            return None
    return val

def regex_find_string(pattern: str, data: str) -> Optional[str]:
    pattern = pattern.replace(r"(%S)", r"(.*)")
    match = re.search(pattern, data)
    if match:
        return match.group(1).strip('"')
    return None

def regex_find_min_float(pattern: str, data: str) -> Optional[float]:
    result = regex_find_floats(pattern, data)
    return min(result) if result else None

def regex_find_max_float(pattern: str, data: str) -> Optional[float]:
    result = regex_find_floats(pattern, data)
    return max(result) if result else None


# Slicer parsing implementations
class BaseSlicer(object):
    def __init__(self, file_path: str) -> None:
        self.path = file_path
        self.header_data: str = ""
        self.footer_data: str = ""
        self.layer_height: Optional[float] = None
        self.has_m486_objects: bool = False

    def set_data(self,
                 header_data: str,
                 footer_data: str,
                 fsize: int) -> None:
        self.header_data = header_data
        self.footer_data = footer_data
        self.size: int = fsize

    def _check_has_objects(self,
                           data: str,
                           pattern: Optional[str] = None
                           ) -> bool:
        match = re.search(
            r"\n((DEFINE_OBJECT)|(EXCLUDE_OBJECT_DEFINE)) NAME=",
            data
        )
        if match is not None:
            # Objects already processed
            fname = os.path.basename(self.path)
            logger.info(
                f"File '{fname}' currently supports cancellation, "
                "processing aborted"
            )
            if match.group(1).startswith("DEFINE_OBJECT"):
                logger.info(
                    "Legacy object processing detected.  This is not "
                    "compatible with official versions of Klipper."
                )
            return False
        # Always check M486
        patterns = [r"\nM486"]
        if pattern is not None:
            patterns.append(pattern)
        for regex in patterns:
            if re.search(regex, data) is not None:
                self.has_m486_objects = regex == r"\nM486"
                return True
        return False

    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        return None

    def has_objects(self) -> bool:
        return self._check_has_objects(self.header_data)

    def parse_gcode_start_byte(self) -> Optional[int]:
        m = re.search(r"\n[MG]\d+\s.*\n", self.header_data)
        if m is None:
            return None
        return m.start()

    def parse_gcode_end_byte(self) -> Optional[int]:
        rev_data = self.footer_data[::-1]
        m = re.search(r"\n.*\s\d+[MG]\n", rev_data)
        if m is None:
            return None
        return self.size - m.start()

    def parse_first_layer_height(self) -> Optional[float]:
        return None

    def parse_layer_height(self) -> Optional[float]:
        return None

    def parse_object_height(self) -> Optional[float]:
        return None

    def parse_filament_total(self) -> Optional[float]:
        return None

    def parse_filament_weight_total(self) -> Optional[float]:
        return None

    def parse_filament_name(self) -> Optional[str]:
        return None

    def parse_filament_type(self) -> Optional[str]:
        return None

    def parse_estimated_time(self) -> Optional[float]:
        return None

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return None

    def parse_chamber_temp(self) -> Optional[float]:
        return None

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return None

    def parse_thumbnails(self) -> Optional[List[Dict[str, Any]]]:
        for data in [self.header_data, self.footer_data]:
            thumb_matches: List[str] = re.findall(
                r"; thumbnail begin[;/\+=\w\s]+?; thumbnail end", data)
            if thumb_matches:
                break
        else:
            return None
        thumb_dir = os.path.join(os.path.dirname(self.path), ".thumbs")
        if not os.path.exists(thumb_dir):
            try:
                os.mkdir(thumb_dir)
            except Exception:
                logger.info(f"Unable to create thumb dir: {thumb_dir}")
                return None
        thumb_base = os.path.splitext(os.path.basename(self.path))[0]
        parsed_matches: List[Dict[str, Any]] = []
        has_miniature: bool = False
        for match in thumb_matches:
            lines = re.split(r"\r?\n", match.replace('; ', ''))
            info = regex_find_ints(r"(%D)", lines[0])
            data = "".join(lines[1:-1])
            if len(info) != 3:
                logger.info(
                    f"MetadataError: Error parsing thumbnail"
                    f" header: {lines[0]}")
                continue
            if len(data) != info[2]:
                logger.info(
                    f"MetadataError: Thumbnail Size Mismatch: "
                    f"detected {info[2]}, actual {len(data)}")
                continue
            thumb_name = f"{thumb_base}-{info[0]}x{info[1]}.png"
            thumb_path = os.path.join(thumb_dir, thumb_name)
            rel_thumb_path = os.path.join(".thumbs", thumb_name)
            with open(thumb_path, "wb") as f:
                f.write(base64.b64decode(data.encode()))
            parsed_matches.append({
                'width': info[0], 'height': info[1],
                'size': os.path.getsize(thumb_path),
                'relative_path': rel_thumb_path})
            if info[0] == 32 and info[1] == 32:
                has_miniature = True
        if len(parsed_matches) > 0 and not has_miniature:
            # find the largest thumb index
            largest_match = parsed_matches[0]
            for item in parsed_matches:
                if item['size'] > largest_match['size']:
                    largest_match = item
            # Create miniature thumbnail if one does not exist
            thumb_full_name = largest_match['relative_path'].split("/")[-1]
            thumb_path = os.path.join(thumb_dir, f"{thumb_full_name}")
            rel_path_small = os.path.join(".thumbs", f"{thumb_base}-32x32.png")
            thumb_path_small = os.path.join(
                thumb_dir, f"{thumb_base}-32x32.png")
            # read file
            try:
                with Image.open(thumb_path) as im:
                    # Create 32x32 thumbnail
                    im.thumbnail((32, 32))
                    im.save(thumb_path_small, format="PNG")
                    parsed_matches.insert(0, {
                        'width': im.width, 'height': im.height,
                        'size': os.path.getsize(thumb_path_small),
                        'relative_path': rel_path_small
                    })
            except Exception as e:
                logger.info(str(e))
        return parsed_matches

    def parse_layer_count(self) -> Optional[int]:
        return None

    def parse_nozzle_diameter(self) -> Optional[float]:
        return None

class UnknownSlicer(BaseSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        return {'slicer': "Unknown"}

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_min_float(r"G1\sZ(%F)\s", self.header_data)

    def parse_object_height(self) -> Optional[float]:
        return regex_find_max_float(r"G1\sZ(%F)\s", self.footer_data)

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(r"M109 S(%F)", self.header_data)

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(r"M190 S(%F)", self.header_data)

    def parse_chamber_temp(self) -> Optional[float]:
        return regex_find_float(r"M191 S(%F)", self.header_data)

    def parse_thumbnails(self) -> Optional[List[Dict[str, Any]]]:
        return None

class PrusaSlicer(BaseSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        aliases = {
            'PrusaSlicer': r"PrusaSlicer\s(.*)\son",
            'SuperSlicer': r"SuperSlicer\s(.*)\son",
            'OrcaSlicer': r"OrcaSlicer\s(.*)\son",
            'MomentSlicer': r"MomentSlicer\s(.*)\son",
            'SliCR-3D': r"SliCR-3D\s(.*)\son",
            'BambuStudio': r"BambuStudio[^ ]*\s(.*)\n",
            'A3dp-Slicer': r"A3dp-Slicer\s(.*)\son",
            'QIDISlicer': r"QIDISlicer\s(.*)\son",
        }
        for name, expr in aliases.items():
            match = re.search(expr, data)
            if match:
                return {
                    'slicer': name,
                    'slicer_version': match.group(1)
                }
        return None

    def has_objects(self) -> bool:
        return self._check_has_objects(
            self.header_data, r"\n; printing object")

    def parse_first_layer_height(self) -> Optional[float]:
        # Check percentage
        pct = regex_find_float(r"; first_layer_height = (%F)%", self.footer_data)
        if pct is not None:
            if self.layer_height is None:
                # Failed to parse the original layer height, so it is not
                # possible to calculate a percentage
                return None
            return round(pct / 100. * self.layer_height, 6)
        return regex_find_float(r"; first_layer_height = (%F)", self.footer_data)

    def parse_layer_height(self) -> Optional[float]:
        self.layer_height = regex_find_float(
            r"; layer_height = (%F)", self.footer_data
        )
        return self.layer_height

    def parse_object_height(self) -> Optional[float]:
        matches = re.findall(
            r";BEFORE_LAYER_CHANGE\n(?:.*\n)?;(\d+\.?\d*)", self.footer_data)
        if matches:
            try:
                matches = [float(m) for m in matches]
            except Exception:
                pass
            else:
                return max(matches)
        return regex_find_max_float(r"G1\sZ(%F)\sF", self.footer_data)

    def parse_filament_total(self) -> Optional[float]:
        line = regex_find_string(r'filament\sused\s\[mm\]\s=\s(%S)\n', self.footer_data)
        if line:
            filament = regex_find_floats(
                r"(%F)", line
            )
            if filament:
                return sum(filament)
        return None

    def parse_filament_weight_total(self) -> Optional[float]:
        return regex_find_float(
            r"total\sfilament\sused\s\[g\]\s=\s(%F)",
            self.footer_data
        )

    def parse_filament_type(self) -> Optional[str]:
        return regex_find_string(r";\sfilament_type\s=\s(%S)", self.footer_data)

    def parse_filament_name(self) -> Optional[str]:
        return regex_find_string(
            r";\sfilament_settings_id\s=\s(%S)", self.footer_data
        )

    def parse_estimated_time(self) -> Optional[float]:
        time_match = re.search(
            r';\sestimated\sprinting\stime.*', self.footer_data)
        if not time_match:
            return None
        total_time = 0
        time_group = time_match.group()
        time_patterns = [(r"(\d+)d", 24*60*60), (r"(\d+)h", 60*60),
                         (r"(\d+)m", 60), (r"(\d+)s", 1)]
        try:
            for pattern, multiplier in time_patterns:
                t = re.search(pattern, time_group)
                if t:
                    total_time += int(t.group(1)) * multiplier
        except Exception:
            return None
        return round(total_time, 2)

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(
            r"; first_layer_temperature = (%F)", self.footer_data
        )

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(
            r"; first_layer_bed_temperature = (%F)", self.footer_data
        )

    def parse_chamber_temp(self) -> Optional[float]:
        return regex_find_float(
            r"; chamber_temperature = (%F)", self.footer_data
        )

    def parse_nozzle_diameter(self) -> Optional[float]:
        return regex_find_float(
            r";\snozzle_diameter\s=\s(%F)", self.footer_data
        )

    def parse_layer_count(self) -> Optional[int]:
        return regex_find_int(r"; total layers count = (%D)", self.footer_data)

class Slic3rPE(PrusaSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        match = re.search(r"Slic3r\sPrusa\sEdition\s(.*)\son", data)
        if match:
            return {
                'slicer': "Slic3r PE",
                'slicer_version': match.group(1)
            }
        return None

    def parse_filament_total(self) -> Optional[float]:
        return regex_find_float(r"filament\sused\s=\s(%F)mm", self.footer_data)

    def parse_thumbnails(self) -> Optional[List[Dict[str, Any]]]:
        return None

class Slic3r(Slic3rPE):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        match = re.search(r"Slic3r\s(\d.*)\son", data)
        if match:
            return {
                'slicer': "Slic3r",
                'slicer_version': match.group(1)
            }
        return None

    def parse_filament_total(self) -> Optional[float]:
        filament = regex_find_float(
            r";\sfilament\_length\_m\s=\s(%F)", self.footer_data
        )
        if filament is not None:
            filament *= 1000
        return filament

    def parse_filament_weight_total(self) -> Optional[float]:
        return regex_find_float(r";\sfilament\smass\_g\s=\s(%F)", self.footer_data)

    def parse_estimated_time(self) -> Optional[float]:
        return None

class Cura(BaseSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        match = re.search(r"Cura_SteamEngine\s(.*)", data)
        if match:
            return {
                'slicer': "Cura",
                'slicer_version': match.group(1)
            }
        return None

    def has_objects(self) -> bool:
        return self._check_has_objects(self.header_data, r"\n;MESH:")

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_float(r";MINZ:(%F)", self.header_data)

    def parse_layer_height(self) -> Optional[float]:
        self.layer_height = regex_find_float(
            r";Layer\sheight:\s(%F)", self.header_data
        )
        return self.layer_height

    def parse_object_height(self) -> Optional[float]:
        return regex_find_float(r";MAXZ:(%F)", self.header_data)

    def parse_filament_total(self) -> Optional[float]:
        filament = regex_find_float(r";Filament\sused:\s(%F)m", self.header_data)
        if filament is not None:
            filament *= 1000
        return filament

    def parse_filament_weight_total(self) -> Optional[float]:
        return regex_find_float(r";Filament\sweight\s=\s.(%F).", self.header_data)

    def parse_filament_type(self) -> Optional[str]:
        return regex_find_string(r";Filament\stype\s=\s(%S)", self.header_data)

    def parse_filament_name(self) -> Optional[str]:
        return regex_find_string(r";Filament\sname\s=\s(%S)", self.header_data)

    def parse_estimated_time(self) -> Optional[float]:
        return regex_find_max_float(r";TIME:(%F)", self.header_data)

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(r"M109 S(%F)", self.header_data)

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(r"M190 S(%F)", self.header_data)

    def parse_chamber_temp(self) -> Optional[float]:
        return regex_find_float(r"M191 S(%F)", self.header_data)

    def parse_layer_count(self) -> Optional[int]:
        return regex_find_int(r";LAYER_COUNT\:(%D)", self.header_data)

    def parse_nozzle_diameter(self) -> Optional[float]:
        return regex_find_float(r";Nozzle\sdiameter\s=\s(%F)", self.header_data)

    def parse_thumbnails(self) -> Optional[List[Dict[str, Any]]]:
        # Attempt to parse thumbnails from file metadata
        thumbs = super().parse_thumbnails()
        if thumbs is not None:
            return thumbs
        # Check for thumbnails extracted from the ufp
        thumb_dir = os.path.join(os.path.dirname(self.path), ".thumbs")
        thumb_base = os.path.splitext(os.path.basename(self.path))[0]
        thumb_path = os.path.join(thumb_dir, f"{thumb_base}.png")
        rel_path_full = os.path.join(".thumbs", f"{thumb_base}.png")
        rel_path_small = os.path.join(".thumbs", f"{thumb_base}-32x32.png")
        thumb_path_small = os.path.join(thumb_dir, f"{thumb_base}-32x32.png")
        if not os.path.isfile(thumb_path):
            return None
        # read file
        thumbs = []
        try:
            with Image.open(thumb_path) as im:
                thumbs.append({
                    'width': im.width, 'height': im.height,
                    'size': os.path.getsize(thumb_path),
                    'relative_path': rel_path_full
                })
                # Create 32x32 thumbnail
                im.thumbnail((32, 32), Image.Resampling.LANCZOS)
                im.save(thumb_path_small, format="PNG")
                thumbs.insert(0, {
                    'width': im.width, 'height': im.height,
                    'size': os.path.getsize(thumb_path_small),
                    'relative_path': rel_path_small
                })
        except Exception as e:
            logger.info(str(e))
            return None
        return thumbs

class Simplify3D(BaseSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        match = re.search(r"Simplify3D\(R\)\sVersion\s(.*)", data)
        if match:
            self._version = match.group(1)
            self._is_v5 = self._version.startswith("5")
            return {
                'slicer': "Simplify3D",
                'slicer_version': match.group(1)
            }
        return None

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_min_float(r"G1\sZ(%F)\s", self.header_data)

    def parse_layer_height(self) -> Optional[float]:
        self.layer_height = regex_find_float(
            r";\s+layerHeight,(%F)", self.header_data
        )
        return self.layer_height

    def parse_object_height(self) -> Optional[float]:
        return regex_find_max_float(r"G1\sZ(%F)\s", self.footer_data)

    def parse_filament_total(self) -> Optional[float]:
        return regex_find_float(
            r";\s+(?:Filament\slength|Material\sLength):\s(%F)\smm",
            self.footer_data
        )

    def parse_filament_weight_total(self) -> Optional[float]:
        return regex_find_float(
            r";\s+(?:Plastic\sweight|Material\sWeight):\s(%F)\sg",
            self.footer_data
        )

    def parse_filament_name(self) -> Optional[str]:
        return regex_find_string(
            r";\s+printMaterial,(%S)", self.header_data)

    def parse_filament_type(self) -> Optional[str]:
        return regex_find_string(
            r";\s+makerBotModelMaterial,(%S)", self.footer_data)

    def parse_estimated_time(self) -> Optional[float]:
        time_match = re.search(r';\s+Build (t|T)ime:.*', self.footer_data)
        if not time_match:
            return None
        total_time = 0
        time_group = time_match.group()
        time_patterns = [(r"(\d+)\shours?", 60*60), (r"(\d+)\smin", 60),
                         (r"(\d+)\ssec", 1)]
        try:
            for pattern, multiplier in time_patterns:
                t = re.search(pattern, time_group)
                if t:
                    total_time += int(t.group(1)) * multiplier
        except Exception:
            return None
        return round(total_time, 2)

    def _get_temp_items(self, pattern: str) -> List[str]:
        match = re.search(pattern, self.header_data)
        if match is None:
            return []
        return match.group().split(",")[1:]

    def _get_first_layer_temp(self, heater: str) -> Optional[float]:
        heaters = self._get_temp_items(r"temperatureName.*")
        temps = self._get_temp_items(r"temperatureSetpointTemperatures.*")
        for h, temp in zip(heaters, temps):
            if h == heater:
                try:
                    return float(temp)
                except Exception:
                    return None
        return None

    def _get_first_layer_temp_v5(self, heater_type: str) -> Optional[float]:
        pattern = (
            r";\s+temperatureController,.+?"
            r";\s+temperatureType,"f"{heater_type}"r".+?"
            r";\s+temperatureSetpoints,\d+\|(\d+)"
        )
        match = re.search(pattern, self.header_data, re.MULTILINE | re.DOTALL)
        if match is not None:
            try:
                return float(match.group(1))
            except Exception:
                return None
        return None

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        if self._is_v5:
            return self._get_first_layer_temp_v5("extruder")
        else:
            return self._get_first_layer_temp("Extruder 1")

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        if self._is_v5:
            return self._get_first_layer_temp_v5("platform")
        else:
            return self._get_first_layer_temp("Heated Bed")

    def parse_nozzle_diameter(self) -> Optional[float]:
        return regex_find_float(
            r";\s+(?:extruderDiameter|nozzleDiameter),(%F)",
            self.header_data
        )

class KISSlicer(BaseSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, Any]]:
        match = re.search(r";\sKISSlicer", data)
        if match:
            ident = {'slicer': "KISSlicer"}
            vmatch = re.search(r";\sversion\s(.*)", data)
            if vmatch:
                version = vmatch.group(1).replace(" ", "-")
                ident['slicer_version'] = version
            return ident
        return None

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_float(
            r";\s+first_layer_thickness_mm\s=\s(%F)", self.header_data)

    def parse_layer_height(self) -> Optional[float]:
        self.layer_height = regex_find_float(
            r";\s+max_layer_thickness_mm\s=\s(%F)", self.header_data)
        return self.layer_height

    def parse_object_height(self) -> Optional[float]:
        return regex_find_max_float(
            r";\sEND_LAYER_OBJECT\sz=(%F)", self.footer_data)

    def parse_filament_total(self) -> Optional[float]:
        filament = regex_find_floats(
            r";\s+Ext #\d+\s+=\s+(%F)\s*mm", self.footer_data)
        if filament:
            return sum(filament)
        return None

    def parse_estimated_time(self) -> Optional[float]:
        time = regex_find_float(
            r";\sCalculated.*Build\sTime:\s(%F)\sminutes",
            self.footer_data)
        if time is not None:
            time *= 60
            return round(time, 2)
        return None

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(r"; first_layer_C = (%F)", self.header_data)

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(r"; bed_C = (%F)", self.header_data)

    def parse_chamber_temp(self) -> Optional[float]:
        return regex_find_float(r"; chamber_C = (%F)", self.header_data)


class IdeaMaker(BaseSlicer):
    def check_identity(self, data: str) -> Optional[Dict[str, str]]:
        match = re.search(r"\sideaMaker\s(.*),", data)
        if match:
            return {
                'slicer': "IdeaMaker",
                'slicer_version': match.group(1)
            }
        return None

    def has_objects(self) -> bool:
        return self._check_has_objects(self.header_data, r"\n;PRINTING:")

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_float(
            r";LAYER:0\s*.*\s*;HEIGHT:(%F)", self.header_data
        )

    def parse_layer_height(self) -> Optional[float]:
        return regex_find_float(
            r";LAYER:1\s*.*\s*;HEIGHT:(%F)", self.header_data
        )

    def parse_object_height(self) -> Optional[float]:
        return regex_find_float(r";Bounding Box:(?:\s+(%F))+", self.header_data)

    def parse_filament_total(self) -> Optional[float]:
        filament = regex_find_floats(
            r";Material.\d\sUsed:\s+(%F)", self.footer_data
        )
        if filament:
            return sum(filament)
        return None

    def parse_filament_type(self) -> Optional[str]:
        return (
            regex_find_string(r";Filament\sType\s.\d:\s(%S)", self.header_data) or
            regex_find_string(r";Filament\stype\s=\s(%S)", self.header_data)
        )

    def parse_filament_name(self) -> Optional[str]:
        return (
            regex_find_string(r";Filament\sName\s.\d:\s(%S)", self.header_data) or
            regex_find_string(r";Filament\sname\s=\s(%S)", self.header_data)
        )

    def parse_filament_weight_total(self) -> Optional[float]:
        pi = 3.141592653589793
        length = regex_find_floats(
            r";Material.\d\sUsed:\s+(%F)", self.footer_data)
        diameter = regex_find_floats(
            r";Filament\sDiameter\s.\d:\s+(%F)", self.header_data)
        density = regex_find_floats(
            r";Filament\sDensity\s.\d:\s+(%F)", self.header_data)
        if len(length) == len(density) == len(diameter):
            # calc individual weight for each filament with m=pi/4*d²*l*rho
            weights = [(pi/4 * diameter[i]**2 * length[i] * density[i]/10**6)
                       for i in range(len(length))]
            return sum(weights)
        return None

    def parse_estimated_time(self) -> Optional[float]:
        return regex_find_float(r";Print\sTime:\s(%F)", self.footer_data)

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(r"M109 T0 S(%F)", self.header_data)

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(r"M190 S(%F)", self.header_data)

    def parse_chamber_temp(self) -> Optional[float]:
        return regex_find_float(r"M191 S(%F)", self.header_data)

    def parse_nozzle_diameter(self) -> Optional[float]:
        return regex_find_float(
            r";Dimension:(?:\s\d+\.\d+){3}\s(%F)", self.header_data)

class IceSL(BaseSlicer):
    def check_identity(self, data) -> Optional[Dict[str, Any]]:
        match = re.search(r"<IceSL\s(.*)>", data)
        if match:
            version = match.group(1) if match.group(1)[0].isdigit() else "-"
            return {
                'slicer': "IceSL",
                'slicer_version': version
            }
        return None

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_float(
            r";\sz_layer_height_first_layer_mm\s:\s+(%F)",
            self.header_data)

    def parse_layer_height(self) -> Optional[float]:
        self.layer_height = regex_find_float(
            r";\sz_layer_height_mm\s:\s+(%F)",
            self.header_data)
        return self.layer_height

    def parse_object_height(self) -> Optional[float]:
        return regex_find_float(
            r";\sprint_height_mm\s:\s+(%F)", self.header_data)

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(
            r";\sextruder_temp_degree_c_0\s:\s+(%F)", self.header_data)

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(
            r";\sbed_temp_degree_c\s:\s+(%F)", self.header_data)

    def parse_chamber_temp(self) -> Optional[float]:
        return regex_find_float(
            r";\schamber_temp_degree_c\s:\s+(%F)", self.header_data)

    def parse_filament_total(self) -> Optional[float]:
        return regex_find_float(
            r";\sfilament_used_mm\s:\s+(%F)", self.header_data)

    def parse_filament_weight_total(self) -> Optional[float]:
        return regex_find_float(
            r";\sfilament_used_g\s:\s+(%F)", self.header_data)

    def parse_filament_name(self) -> Optional[str]:
        return regex_find_string(
            r";\sfilament_name\s:\s+(%S)", self.header_data)

    def parse_filament_type(self) -> Optional[str]:
        return regex_find_string(
            r";\sfilament_type\s:\s+(%S)", self.header_data)

    def parse_estimated_time(self) -> Optional[float]:
        return regex_find_float(
            r";\sestimated_print_time_s\s:\s+(%F)", self.header_data)

    def parse_layer_count(self) -> Optional[int]:
        return regex_find_int(
            r";\slayer_count\s:\s+(%D)", self.header_data)

    def parse_nozzle_diameter(self) -> Optional[float]:
        return regex_find_float(
            r";\snozzle_diameter_mm_0\s:\s+(%F)", self.header_data)

class KiriMoto(BaseSlicer):
    def check_identity(self, data) -> Optional[Dict[str, Any]]:
        variants: Dict[str, str] = {
            "Kiri:Moto": r"; Generated by Kiri:Moto (\d.+)",
            "SimplyPrint": r"; Generated by Kiri:Moto \(SimplyPrint\) (.+)"
        }
        for name, pattern in variants.items():
            match = re.search(pattern, data)
            if match:
                return {
                    "slicer": name,
                    "slicer_version": match.group(1)
                }
        return None

    def parse_first_layer_height(self) -> Optional[float]:
        return regex_find_float(
            r"; firstSliceHeight = (%F)", self.header_data
        )

    def parse_layer_height(self) -> Optional[float]:
        self.layer_height = regex_find_float(
            r"; sliceHeight = (%F)", self.header_data
        )
        return self.layer_height

    def parse_object_height(self) -> Optional[float]:
        return regex_find_max_float(
            r"G1 Z(%F) (?:; z-hop end|F\d+\n)", self.footer_data
        )

    def parse_layer_count(self) -> Optional[int]:
        matches = re.findall(
            r";; --- layer (\d+) \(.+", self.footer_data
        )
        if not matches:
            return None
        try:
            return int(matches[-1]) + 1
        except Exception:
            return None

    def parse_estimated_time(self) -> Optional[float]:
        return regex_find_int(r"; --- print time: (%D)s", self.footer_data)

    def parse_filament_total(self) -> Optional[float]:
        return regex_find_float(
            r"; --- filament used: (%F) mm", self.footer_data
        )

    def parse_first_layer_extr_temp(self) -> Optional[float]:
        return regex_find_float(
            r"; firstLayerNozzleTemp = (%F)", self.header_data
        )

    def parse_first_layer_bed_temp(self) -> Optional[float]:
        return regex_find_float(
            r"; firstLayerBedTemp = (%F)", self.header_data
        )


READ_SIZE = 1024 * 1024  # 1 MiB
SUPPORTED_SLICERS: List[Type[BaseSlicer]] = [
    PrusaSlicer, Slic3rPE, Slic3r, Cura, Simplify3D,
    KISSlicer, IdeaMaker, IceSL, KiriMoto
]
SUPPORTED_DATA = [
    'gcode_start_byte',
    'gcode_end_byte',
    'layer_count',
    'object_height',
    'estimated_time',
    'nozzle_diameter',
    'layer_height',
    'first_layer_height',
    'first_layer_extr_temp',
    'first_layer_bed_temp',
    'chamber_temp',
    'filament_name',
    'filament_type',
    'filament_total',
    'filament_weight_total',
    'thumbnails']

def process_objects(file_path: str, slicer: BaseSlicer, name: str) -> bool:
    try:
        from preprocess_cancellation import (
            preprocess_slicer,
            preprocess_cura,
            preprocess_ideamaker,
            preprocess_m486
        )
    except ImportError:
        logger.info("Module 'preprocess-cancellation' failed to load")
        return False
    fname = os.path.basename(file_path)
    logger.info(
        f"Performing Object Processing on file: {fname}, "
        f"sliced by {name}"
    )
    with tempfile.TemporaryDirectory() as tmp_dir_name:
        tmp_file = os.path.join(tmp_dir_name, fname)
        with open(file_path, 'r') as in_file:
            with open(tmp_file, 'w') as out_file:
                try:
                    if slicer.has_m486_objects:
                        processor = preprocess_m486
                    elif isinstance(slicer, PrusaSlicer):
                        processor = preprocess_slicer
                    elif isinstance(slicer, Cura):
                        processor = preprocess_cura
                    elif isinstance(slicer, IdeaMaker):
                        processor = preprocess_ideamaker
                    else:
                        logger.info(
                            f"Object Processing Failed, slicer {name}"
                            "not supported"
                        )
                        return False
                    for line in processor(in_file):
                        out_file.write(line)
                except Exception as e:
                    logger.info(f"Object processing failed: {e}")
                    return False
        if os.path.islink(file_path):
            file_path = os.path.realpath(file_path)
        shutil.move(tmp_file, file_path)
    return True

def get_slicer(file_path: str) -> Tuple[BaseSlicer, Dict[str, str]]:
    header_data = footer_data = ""
    slicer: Optional[BaseSlicer] = None
    size = os.path.getsize(file_path)
    with open(file_path, 'r') as f:
        # read the default size, which should be enough to
        # identify the slicer
        header_data = f.read(READ_SIZE)
        for impl in SUPPORTED_SLICERS:
            slicer = impl(file_path)
            ident = slicer.check_identity(header_data)
            if ident is not None:
                break
        else:
            slicer = UnknownSlicer(file_path)
            ident = slicer.check_identity(header_data)
        if size > READ_SIZE * 2:
            f.seek(size - READ_SIZE)
            footer_data = f.read()
        elif size > READ_SIZE:
            remaining = size - READ_SIZE
            footer_data = header_data[remaining - READ_SIZE:] + f.read()
        else:
            footer_data = header_data
        slicer.set_data(header_data, footer_data, size)
    if ident is None:
        ident = {"slicer": "unknown"}
    return slicer, ident

def extract_metadata(
    file_path: str, check_objects: bool
) -> Dict[str, Any]:
    metadata: Dict[str, Any] = {}
    slicer, ident = get_slicer(file_path)
    if check_objects and slicer.has_objects():
        name = ident.get("slicer", "unknown")
        if process_objects(file_path, slicer, name):
            slicer, ident = get_slicer(file_path)
    metadata['size'] = os.path.getsize(file_path)
    metadata['modified'] = os.path.getmtime(file_path)
    metadata['uuid'] = str(uuid.uuid4())
    metadata.update(ident)
    for key in SUPPORTED_DATA:
        func = getattr(slicer, "parse_" + key)
        result = func()
        if result is not None:
            metadata[key] = result
    return metadata

def extract_ufp(ufp_path: str, dest_path: str) -> None:
    if not os.path.isfile(ufp_path):
        logger.info(f"UFP file Not Found: {ufp_path}")
        sys.exit(-1)
    thumb_name = os.path.splitext(
        os.path.basename(dest_path))[0] + ".png"
    dest_thumb_dir = os.path.join(os.path.dirname(dest_path), ".thumbs")
    dest_thumb_path = os.path.join(dest_thumb_dir, thumb_name)
    try:
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            tmp_thumb_path = ""
            with zipfile.ZipFile(ufp_path) as zf:
                tmp_model_path = zf.extract(
                    UFP_MODEL_PATH, path=tmp_dir_name)
                if UFP_THUMB_PATH in zf.namelist():
                    tmp_thumb_path = zf.extract(
                        UFP_THUMB_PATH, path=tmp_dir_name)
            if os.path.islink(dest_path):
                dest_path = os.path.realpath(dest_path)
            shutil.move(tmp_model_path, dest_path)
            if tmp_thumb_path:
                if not os.path.exists(dest_thumb_dir):
                    os.mkdir(dest_thumb_dir)
                shutil.move(tmp_thumb_path, dest_thumb_path)
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(-1)
    try:
        os.remove(ufp_path)
    except Exception:
        logger.info(f"Error removing ufp file: {ufp_path}")

def main(path: str,
         filename: str,
         ufp: Optional[str],
         check_objects: bool
         ) -> None:
    file_path = os.path.join(path, filename)
    if ufp is not None:
        extract_ufp(ufp, file_path)
    metadata: Dict[str, Any] = {}
    if not os.path.isfile(file_path):
        logger.info(f"File Not Found: {file_path}")
        sys.exit(-1)
    try:
        metadata = extract_metadata(file_path, check_objects)
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(-1)
    fd = sys.stdout.fileno()
    data = json.dumps(
        {'file': filename, 'metadata': metadata}).encode()
    while data:
        try:
            ret = os.write(fd, data)
        except OSError:
            continue
        data = data[ret:]


if __name__ == "__main__":
    # Parse start arguments
    parser = argparse.ArgumentParser(
        description="GCode Metadata Extraction Utility")
    parser.add_argument(
        "-f", "--filename", metavar='<filename>',
        help="name gcode file to parse")
    parser.add_argument(
        "-p", "--path", default=os.path.abspath(os.path.dirname(__file__)),
        metavar='<path>',
        help="optional absolute path for file"
    )
    parser.add_argument(
        "-u", "--ufp", metavar="<ufp file>", default=None,
        help="optional path of ufp file to extract"
    )
    parser.add_argument(
        "-o", "--check-objects", dest='check_objects', action='store_true',
        help="process gcode file for exclude opbject functionality")
    args = parser.parse_args()
    check_objects = args.check_objects
    enabled_msg = "enabled" if check_objects else "disabled"
    logger.info(f"Object Processing is {enabled_msg}")
    main(args.path, args.filename, args.ufp, check_objects)



# Support for GPIO Button actions
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import asyncio
import logging

from typing import (
    TYPE_CHECKING,
    Any,
    Dict
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from .application import InternalTransport as ITransport


class ButtonManager:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.buttons: Dict[str, GpioButton] = {}
        prefix_sections = config.get_prefix_sections("button")
        logging.info(f"Loading Buttons: {prefix_sections}")
        for section in prefix_sections:
            cfg = config[section]
            # Reserve the "type" option for future use
            btn_type = cfg.get('type', "gpio")  # noqa: F841
            try:
                btn = GpioButton(cfg)
            except Exception as e:
                msg = f"Failed to load button [{cfg.get_name()}]\n{e}"
                self.server.add_warning(msg, exc_info=e)
                continue
            self.buttons[btn.name] = btn
        self.server.register_notification("button:button_event")

    def component_init(self) -> None:
        for btn in self.buttons.values():
            btn.initialize()

class GpioButton:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.eventloop = self.server.get_event_loop()
        self.name = config.get_name().split()[-1]
        self.itransport: ITransport = self.server.lookup_component("internal_transport")
        self.mutex = asyncio.Lock()
        self.gpio_event = config.getgpioevent("pin", self._on_gpio_event)
        self.min_event_time = config.getfloat("minimum_event_time", 0, minval=0.0)
        debounce_period = config.getfloat("debounce_period", .05, minval=0.01)
        self.gpio_event.setup_debounce(debounce_period, self._on_gpio_error)
        self.press_template = config.gettemplate("on_press", None, is_async=True)
        self.release_template = config.gettemplate("on_release", None, is_async=True)
        if (
            self.press_template is None and
            self.release_template is None
        ):
            raise config.error(
                f"[{config.get_name()}]: No template option configured"
            )
        self.notification_sent: bool = False
        self.user_data: Dict[str, Any] = {}
        self.context: Dict[str, Any] = {
            'call_method': self.itransport.call_method,
            'send_notification': self._send_notification,
            'event': {
                'elapsed_time': 0.,
                'received_time': 0.,
                'render_time': 0.,
                'pressed': False,
            },
            'user_data': self.user_data
        }

    def initialize(self) -> None:
        self.gpio_event.start()
        self.context['event']['pressed'] = bool(self.gpio_event.get_value())

    def get_status(self) -> Dict[str, Any]:
        return {
            'name': self.name,
            'type': "gpio",
            'event': self.context['event'],
        }

    def _send_notification(self, result: Any = None) -> None:
        if self.notification_sent:
            # Only allow execution once per template
            return
        self.notification_sent = True
        data = self.get_status()
        data['aux'] = result
        self.server.send_event("button:button_event", data)

    async def _on_gpio_event(
        self, eventtime: float, elapsed_time: float, pressed: int
    ) -> None:
        if elapsed_time < self.min_event_time:
            return
        template = self.press_template if pressed else self.release_template
        if template is None:
            return
        async with self.mutex:
            self.notification_sent = False
            event_info: Dict[str, Any] = {
                'elapsed_time': elapsed_time,
                'received_time': eventtime,
                'render_time': self.eventloop.get_loop_time(),
                'pressed': bool(pressed)
            }
            self.context['event'] = event_info
            try:
                await template.render_async(self.context)
            except Exception:
                action = "on_press" if pressed else "on_release"
                logging.exception(
                    f"Button {self.name}: '{action}' template error")

    def _on_gpio_error(self, message: str) -> None:
        self.server.add_warning(f"Button {self.name}: {message}")

def load_component(config: ConfigHelper) -> ButtonManager:
    return ButtonManager(config)



# LDAP authentication for Moonraker
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
# Copyright (C) 2022 Luca Schöneberg <luca-schoeneberg@outlook.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import asyncio
import logging
import ldap3
from ldap3.core.exceptions import LDAPExceptionError

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Optional
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ldap3.abstract.entry import Entry

class MoonrakerLDAP:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.ldap_host = config.get('ldap_host')
        self.ldap_port = config.getint("ldap_port", None)
        self.ldap_secure = config.getboolean("ldap_secure", False)
        base_dn_template = config.gettemplate('base_dn')
        self.base_dn = base_dn_template.render()
        self.group_dn: Optional[str] = None
        group_dn_template = config.gettemplate("group_dn", None)
        if group_dn_template is not None:
            self.group_dn = group_dn_template.render()
        self.active_directory = config.getboolean('is_active_directory', False)
        self.bind_dn: Optional[str] = None
        self.bind_password: Optional[str] = None
        bind_dn_template = config.gettemplate('bind_dn', None)
        bind_pass_template = config.gettemplate('bind_password', None)
        if bind_dn_template is not None:
            self.bind_dn = bind_dn_template.render()
            if bind_pass_template is None:
                raise config.error(
                    "Section [ldap]: Option 'bind_password' is "
                    "required when 'bind_dn' is provided"
                )
            self.bind_password = bind_pass_template.render()
        self.user_filter: Optional[str] = None
        user_filter_template = config.gettemplate('user_filter', None)
        if user_filter_template is not None:
            self.user_filter = user_filter_template.render()
            if "USERNAME" not in self.user_filter:
                raise config.error(
                    "Section [ldap]: Option 'user_filter' is "
                    "is missing required token USERNAME"
                )
        self.lock = asyncio.Lock()

    async def authenticate_ldap_user(self, username, password) -> None:
        eventloop = self.server.get_event_loop()
        async with self.lock:
            await eventloop.run_in_thread(
                self._perform_ldap_auth, username, password
            )

    def _perform_ldap_auth(self, username, password) -> None:
        server = ldap3.Server(
            self.ldap_host, self.ldap_port, use_ssl=self.ldap_secure,
            connect_timeout=10.
        )
        conn_args = {
            "user": self.bind_dn,
            "password": self.bind_password,
            "auto_bind": ldap3.AUTO_BIND_NO_TLS,
        }
        attr_name = "sAMAccountName" if self.active_directory else "uid"
        ldfilt = f"(&(objectClass=Person)({attr_name}={username}))"
        if self.user_filter:
            ldfilt = self.user_filter.replace("USERNAME", username)
        try:
            with ldap3.Connection(server, **conn_args) as conn:
                ret = conn.search(
                    self.base_dn, ldfilt, attributes=["memberOf"]
                )
                if not ret:
                    raise self.server.error(
                        f"LDAP User '{username}' Not Found", 401
                    )
                user: Entry = conn.entries[0]
                rebind_success = conn.rebind(user.entry_dn, password)
            if not rebind_success:
                # Server may not allow rebinding, attempt to start
                # a new connection to validate credentials
                logging.debug(
                    "LDAP Rebind failed, attempting to validate credentials "
                    "with new connection."
                )
                conn_args["user"] = user.entry_dn
                conn_args["password"] = password
                with ldap3.Connection(server, **conn_args) as conn:
                    if self._validate_group(username, user):
                        return
            elif self._validate_group(username, user):
                return
        except LDAPExceptionError:
            err_msg = "LDAP authentication failed"
        else:
            err_msg = "Invalid LDAP Username or Password"
        raise self.server.error(err_msg, 401)

    def _validate_group(self, username: str, user: Entry) -> bool:
        if self.group_dn is None:
            logging.debug(f"LDAP User {username} login successful")
            return True
        if not hasattr(user, "memberOf"):
            return False
        for group in user.memberOf.values:
            if group == self.group_dn:
                logging.debug(
                    f"LDAP User {username} group match success, "
                    "login successful"
                )
                return True
        return False


def load_component(config: ConfigHelper) -> MoonrakerLDAP:
    return MoonrakerLDAP(config)



# Log Management
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import logging
import logging.handlers
import time
import os
import sys
import asyncio
import platform
from queue import SimpleQueue as Queue
from .common import RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Optional,
    Awaitable,
    Dict,
    List,
    Any,
)

if TYPE_CHECKING:
    from .server import Server
    from .common import WebRequest
    from .components.klippy_connection import KlippyConnection

# Coroutine friendly QueueHandler courtesy of Martjin Pieters:
# https://www.zopatista.com/python/2019/05/11/asyncio-logging/
class LocalQueueHandler(logging.handlers.QueueHandler):
    def emit(self, record: logging.LogRecord) -> None:
        # Removed the call to self.prepare(), handle task cancellation
        try:
            self.enqueue(record)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.handleError(record)

# Timed Rotating File Handler, based on Klipper's implementation
class MoonrakerLoggingHandler(logging.handlers.TimedRotatingFileHandler):
    def __init__(self, app_args: Dict[str, Any], **kwargs) -> None:
        super().__init__(app_args['log_file'], **kwargs)
        self.app_args = app_args
        self.rollover_info: Dict[str, str] = {}

    def set_rollover_info(self, name: str, item: str) -> None:
        self.rollover_info[name] = item

    def doRollover(self) -> None:
        super().doRollover()
        self.write_header()

    def write_header(self) -> None:
        if self.stream is None:
            return
        strtime = time.asctime(time.gmtime())
        header = f"{'-'*20} Log Start | {strtime} {'-'*20}\n"
        self.stream.write(header)
        self.stream.write(f"platform: {platform.platform(terse=True)}\n")
        app_section = "\n".join([f"{k}: {v}" for k, v in self.app_args.items()])
        self.stream.write(app_section + "\n")
        if self.rollover_info:
            lines = [line for line in self.rollover_info.values() if line]
            self.stream.write("\n".join(lines) + "\n")

class LogManager:
    def __init__(
        self, app_args: Dict[str, Any], startup_warnings: List[str]
    ) -> None:
        root_logger = logging.getLogger()
        while root_logger.hasHandlers():
            root_logger.removeHandler(root_logger.handlers[0])
        queue: Queue = Queue()
        queue_handler = LocalQueueHandler(queue)
        root_logger.addHandler(queue_handler)
        root_logger.setLevel(logging.INFO)
        stdout_hdlr = logging.StreamHandler(sys.stdout)
        stdout_fmt = logging.Formatter(
            '[%(filename)s:%(funcName)s()] - %(message)s')
        stdout_hdlr.setFormatter(stdout_fmt)
        app_args_str = f"platform: {platform.platform(terse=True)}\n"
        app_args_str += "\n".join([f"{k}: {v}" for k, v in app_args.items()])
        sys.stdout.write(f"\nApplication Info:\n{app_args_str}\n")
        self.file_hdlr: Optional[MoonrakerLoggingHandler] = None
        self.listener: Optional[logging.handlers.QueueListener] = None
        log_file: str = app_args.get('log_file', "")
        if log_file:
            try:
                self.file_hdlr = MoonrakerLoggingHandler(
                    app_args, when='midnight', backupCount=2)
                formatter = logging.Formatter(
                    '%(asctime)s [%(filename)s:%(funcName)s()] - %(message)s')
                self.file_hdlr.setFormatter(formatter)
                self.listener = logging.handlers.QueueListener(
                    queue, self.file_hdlr, stdout_hdlr)
                self.file_hdlr.write_header()
            except Exception:
                log_file = os.path.normpath(log_file)
                dir_name = os.path.dirname(log_file)
                startup_warnings.append(
                    f"Unable to create log file at '{log_file}'. "
                    f"Make sure that the folder '{dir_name}' exists "
                    "and Moonraker has Read/Write access to the folder. "
                )
        if self.listener is None:
            self.listener = logging.handlers.QueueListener(
                queue, stdout_hdlr)
        self.listener.start()

    def set_server(self, server: Server) -> None:
        self.server = server
        self.server.register_endpoint(
            "/server/logs/rollover", RequestType.POST, self._handle_log_rollover
        )

    def set_rollover_info(self, name: str, item: str) -> None:
        if self.file_hdlr is not None:
            self.file_hdlr.set_rollover_info(name, item)

    def rollover_log(self) -> Awaitable[None]:
        if self.file_hdlr is None:
            raise self.server.error("File Logging Disabled")
        eventloop = self.server.get_event_loop()
        return eventloop.run_in_thread(self.file_hdlr.doRollover)

    def stop_logging(self):
        if self.listener is not None:
            self.listener.stop()

    async def _handle_log_rollover(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        log_apps = ["moonraker", "klipper"]
        app = web_request.get_str("application", None)
        result: Dict[str, Any] = {"rolled_over": [], "failed": {}}
        if app is not None:
            if app not in log_apps:
                raise self.server.error(f"Unknown application {app}")
            log_apps = [app]
        if "moonraker" in log_apps:
            try:
                ret = self.rollover_log()
                if ret is not None:
                    await ret
            except asyncio.CancelledError:
                raise
            except Exception as e:
                result["failed"]["moonraker"] = str(e)
            else:
                result["rolled_over"].append("moonraker")
        if "klipper" in log_apps:
            kconn: KlippyConnection
            kconn = self.server.lookup_component("klippy_connection")
            try:
                await kconn.rollover_log()
            except self.server.error as e:
                result["failed"]["klipper"] = str(e)
            else:
                result["rolled_over"].append("klipper")
        return result



# Base Deployment Interface
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import logging
import time
from ...utils import pretty_print_time

from typing import TYPE_CHECKING, Dict, Any, Optional, Coroutine
if TYPE_CHECKING:
    from ...confighelper import ConfigHelper
    from ...utils import ServerError
    from .update_manager import CommandHelper

class BaseDeploy:
    def __init__(self,
                 config: ConfigHelper,
                 cmd_helper: CommandHelper,
                 name: Optional[str] = None,
                 prefix: str = "",
                 cfg_hash: Optional[str] = None
                 ) -> None:
        if name is None:
            name = self.parse_name(config)
        self.name = name
        if prefix:
            prefix = f"{prefix} {self.name}: "
        self.prefix = prefix
        self.server = config.get_server()
        self.cmd_helper = cmd_helper
        self.refresh_interval = cmd_helper.get_refresh_interval()
        refresh_interval = config.getint('refresh_interval', None)
        if refresh_interval is not None:
            self.refresh_interval = refresh_interval * 60 * 60
        if cfg_hash is None:
            cfg_hash = config.get_hash().hexdigest()
        self.cfg_hash = cfg_hash

    @staticmethod
    def parse_name(config: ConfigHelper) -> str:
        name = config.get_name().split(maxsplit=1)[-1]
        if name.startswith("client "):
            # allow deprecated [update_manager client app] style names
            name = name[7:]
        return name

    async def initialize(self) -> Dict[str, Any]:
        umdb = self.cmd_helper.get_umdb()
        storage: Dict[str, Any] = await umdb.get(self.name, {})
        self.last_refresh_time: float = storage.get('last_refresh_time', 0.0)
        self.last_cfg_hash: str = storage.get('last_config_hash', "")
        return storage

    def needs_refresh(self, log_remaining_time: bool = False) -> bool:
        next_refresh_time = self.last_refresh_time + self.refresh_interval
        remaining_time = int(next_refresh_time - time.time() + .5)
        if self.cfg_hash != self.last_cfg_hash or remaining_time <= 0:
            return True
        if log_remaining_time:
            self.log_info(f"Next refresh in: {pretty_print_time(remaining_time)}")
        return False

    def get_last_refresh_time(self) -> float:
        return self.last_refresh_time

    async def refresh(self) -> None:
        pass

    async def update(self) -> bool:
        return False

    async def rollback(self) -> bool:
        raise self.server.error(f"Rollback not available for {self.name}")

    def get_update_status(self) -> Dict[str, Any]:
        return {}

    def get_persistent_data(self) -> Dict[str, Any]:
        return {
            'last_config_hash': self.cfg_hash,
            'last_refresh_time': self.last_refresh_time
        }

    def _save_state(self) -> None:
        umdb = self.cmd_helper.get_umdb()
        self.last_refresh_time = time.time()
        self.last_cfg_hash = self.cfg_hash
        umdb[self.name] = self.get_persistent_data()

    def log_exc(self, msg: str, traceback: bool = True) -> ServerError:
        log_msg = f"{self.prefix}{msg}"
        if traceback:
            logging.exception(log_msg)
        else:
            logging.info(log_msg)
        return self.server.error(msg)

    def log_info(self, msg: str) -> None:
        log_msg = f"{self.prefix}{msg}"
        logging.info(log_msg)

    def log_debug(self, msg: str) -> None:
        log_msg = f"{self.prefix}{msg}"
        logging.debug(log_msg)

    def notify_status(self, msg: str, is_complete: bool = False) -> None:
        log_msg = f"{self.prefix}{msg}"
        logging.debug(log_msg)
        self.cmd_helper.notify_update_response(log_msg, is_complete)

    def close(self) -> Optional[Coroutine]:
        return None



# Package definition for the update_manager
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
from . import update_manager as um

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...confighelper import ConfigHelper

def load_component(config: ConfigHelper) -> um.UpdateManager:
    return um.load_component(config)



#! /usr/bin/python3
# Tool to backup and restore Moonraker's LMDB database
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license
import argparse
import pathlib
import base64
import tempfile
import re
import time
from typing import Any, Dict, Optional, TextIO, Tuple
import lmdb

MAX_NAMESPACES = 100
MAX_DB_SIZE = 200 * 2**20
HEADER_KEY = b"MOONRAKER_DATABASE_START"

LINE_MATCH = re.compile(
    r"^\+(\d+),(\d+):([A-Za-z0-9+/]+={0,2})->([A-Za-z0-9+/]+={0,2})$"
)

class DBToolError(Exception):
    pass

# Use a modified CDBMake Format
# +keylen,datalen:namespace|key->data
# Key length includes the namespace, key and separator (a colon)

def open_db(db_path: str) -> lmdb.Environment:
    return lmdb.open(db_path, map_size=MAX_DB_SIZE,
                     max_dbs=MAX_NAMESPACES)

def _do_dump(namespace: bytes,
             db: object,
             backup: TextIO,
             txn: lmdb.Transaction
             ) -> None:
    expected_key_count: int = txn.stat(db)["entries"]
    # write the namespace header
    ns_key = base64.b64encode(b"namespace_" + namespace).decode()
    ns_str = f"entries={expected_key_count}"
    ns_val = base64.b64encode(ns_str.encode()).decode()
    out = f"+{len(ns_key)},{len(ns_val)}:{ns_key}->{ns_val}\n"
    backup.write(out)
    with txn.cursor(db=db) as cursor:
        count = 0
        remaining = cursor.first()
        while remaining:
            key, value = cursor.item()
            keystr = base64.b64encode(key).decode()
            valstr = base64.b64encode(value).decode()
            out = f"+{len(keystr)},{len(valstr)}:{keystr}->{valstr}\n"
            backup.write(out)
            count += 1
            remaining = cursor.next()
    if expected_key_count != count:
        print("Warning: Key count mismatch for namespace "
              f"'{namespace.decode()}': expected {expected_key_count}"
              f", wrote {count}")

def _write_header(ns_count: int, backup: TextIO):
    val_str = f"namespace_count={ns_count}"
    hkey = base64.b64encode(HEADER_KEY).decode()
    hval = base64.b64encode(val_str.encode()).decode()
    out = f"+{len(hkey)},{len(hval)}:{hkey}->{hval}\n"
    backup.write(out)

def backup(args: Dict[str, Any]):
    source_db = pathlib.Path(args["source"]).expanduser().resolve()
    if not source_db.is_dir():
        print(f"Source path not a folder: '{source_db}'")
        exit(1)
    if not source_db.joinpath("data.mdb").exists():
        print(f"No database file found in source path: '{source_db}'")
        exit(1)
    bkp_dest = pathlib.Path(args["output"]).expanduser().resolve()
    print(f"Backing up database at '{source_db}' to '{bkp_dest}'...")
    if bkp_dest.exists():
        print(f"Warning: file at '{bkp_dest}' exists, will be overwritten")
    env = open_db(str(source_db))
    expected_ns_cnt: int = env.stat()["entries"]
    with bkp_dest.open("wt") as f:
        _write_header(expected_ns_cnt, f)
        with env.begin(buffers=True) as txn:
            count = 0
            with txn.cursor() as cursor:
                remaining = cursor.first()
                while remaining:
                    namespace = bytes(cursor.key())
                    db = env.open_db(namespace, txn=txn, create=False)
                    _do_dump(namespace, db, f, txn)
                    count += 1
                    remaining = cursor.next()
    env.close()
    if expected_ns_cnt != count:
        print("Warning: namespace count mismatch: "
              f"expected: {expected_ns_cnt}, wrote: {count}")
    print("Backup complete!")

def _process_header(key: bytes, value: bytes) -> int:
    if key != HEADER_KEY:
        raise DBToolError(
            "Database Backup does not contain a valid header key, "
            f" got {key.decode()}")
    val_parts = value.split(b"=", 1)
    if val_parts[0] != b"namespace_count":
        raise DBToolError(
            "Database Backup has an invalid header value, got "
            f"{value.decode()}")
    return int(val_parts[1])

def _process_namespace(key: bytes, value: bytes) -> Tuple[bytes, int]:
    key_parts = key.split(b"_", 1)
    if key_parts[0] != b"namespace":
        raise DBToolError(
            f"Invalid Namespace Key '{key.decode()}', ID not prefixed")
    namespace = key_parts[1]
    val_parts = value.split(b"=", 1)
    if val_parts[0] != b"entries":
        raise DBToolError(
            f"Invalid Namespace value '{value.decode()}', entry "
            "count not present")
    entries = int(val_parts[1])
    return namespace, entries

def _process_line(line: str) -> Tuple[bytes, bytes]:
    match = LINE_MATCH.match(line)
    if match is None:
        # TODO: use own exception
        raise DBToolError(
            f"Invalid DB Entry match: {line}")
    parts = match.groups()
    if len(parts) != 4:
        raise DBToolError(
            f"Invalid DB Entry, does not contain all data: {line}")
    key_len, val_len, key, val = parts
    if len(key) != int(key_len):
        raise DBToolError(
            f"Invalid DB Entry, key length mismatch. "
            f"Got {len(key)}, expected {key_len}, line: {line}")
    if len(val) != int(val_len):
        raise DBToolError(
            f"Invalid DB Entry, value length mismatch. "
            f"Got {len(val)}, expected {val_len}, line: {line}")
    decoded_key = base64.b64decode(key.encode())
    decoded_val = base64.b64decode(val.encode())
    return decoded_key, decoded_val

def restore(args: Dict[str, Any]):
    dest_path = pathlib.Path(args["destination"]).expanduser().resolve()
    input_db = pathlib.Path(args["input"]).expanduser().resolve()
    if not input_db.is_file():
        print(f"No backup found at path: {input_db}")
        exit(1)
    if not dest_path.exists():
        print(f"Destination path '{dest_path}' does not exist, directory"
              "will be created")
    print(f"Restoring backup from '{input_db}' to '{dest_path}'...")
    bkp_dir: Optional[pathlib.Path] = None
    if dest_path.joinpath("data.mdb").exists():
        bkp_dir = dest_path.parent.joinpath("backup")
        if not bkp_dir.exists():
            bkp_dir = pathlib.Path(tempfile.gettempdir())
        str_time = time.strftime("%Y%m%dT%H%M%SZ", time.gmtime())
        bkp_dir = bkp_dir.joinpath(f"{str_time}/database")
        if not bkp_dir.is_dir():
            bkp_dir.mkdir(parents=True)
        print(f"Warning: database file at found in '{dest_path}', "
              "all data will be overwritten.  Copying existing DB "
              f"to '{bkp_dir}'")
    env = open_db(str(dest_path))
    if bkp_dir is not None:
        env.copy(str(bkp_dir))
    expected_ns_count = -1
    namespace_count = 0
    keys_left = 0
    namespace = b""
    current_db = object()
    with env.begin(write=True) as txn:
        # clear all existing entries
        dbs = []
        with txn.cursor() as cursor:
            remaining = cursor.first()
            while remaining:
                ns = cursor.key()
                dbs.append(env.open_db(ns, txn=txn, create=False))
                remaining = cursor.next()
        for db in dbs:
            txn.drop(db)
        with input_db.open("rt") as f:
            while True:
                line = f.readline()
                if not line:
                    break
                key, val = _process_line(line)
                if expected_ns_count < 0:
                    expected_ns_count = _process_header(key, val)
                    continue
                if not keys_left:
                    namespace, keys_left = _process_namespace(key, val)
                    current_db = env.open_db(namespace, txn=txn)
                    namespace_count += 1
                    continue
                txn.put(key, val, db=current_db)
                keys_left -= 1
    if expected_ns_count != namespace_count:
        print("Warning: Namespace count mismatch, expected: "
              f"{expected_ns_count}, processed {namespace_count}")
    print("Restore Complete")


if __name__ == "__main__":
    # Parse start arguments
    parser = argparse.ArgumentParser(
        description="dbtool - tool for backup/restore of Moonraker's database")
    subparsers = parser.add_subparsers(
        title="commands", description="valid commands", required=True,
        metavar="<command>")
    bkp_parser = subparsers.add_parser("backup", help="backup a database")
    rst_parser = subparsers.add_parser("restore", help="restore a databse")
    bkp_parser.add_argument(
        "source", metavar="<source path>",
        help="location of the folder containing the database to backup")
    bkp_parser.add_argument(
        "output", metavar="<output file>",
        help="location of the backup file to write to",
        default="~/moonraker_db.bkp")
    bkp_parser.set_defaults(func=backup)
    rst_parser.add_argument(
        "destination", metavar="<destination>",
        help="location of the folder where the database will be restored")
    rst_parser.add_argument(
        "input", metavar="<input file>",
        help="location of the backup file to restore from")
    rst_parser.set_defaults(func=restore)
    args = parser.parse_args()
    args.func(vars(args))



from __future__ import annotations
import asyncio
from utils import ServerError
from .mock_gpio import MockGpiod

__all__ = ("MockReader", "MockWriter", "MockComponent", "MockWebsocket",
           "MockGpiod")

class MockWriter:
    def __init__(self, wait_drain: bool = False) -> None:
        self.wait_drain = wait_drain

    def write(self, data: str) -> None:
        pass

    async def drain(self) -> None:
        if self.wait_drain:
            evt = asyncio.Event()
            await evt.wait()
        else:
            raise ServerError("TestError")

class MockReader:
    def __init__(self, action: str = "") -> None:
        self.action = action
        self.eof = False

    def at_eof(self) -> bool:
        return self.eof

    async def readuntil(self, stop: bytes) -> bytes:
        if self.action == "wait":
            evt = asyncio.Event()
            await evt.wait()
            return b""
        elif self.action == "raise_error":
            raise ServerError("TestError")
        else:
            self.eof = True
            return b"NotJsonDecodable"


class MockComponent:
    def __init__(self,
                 err_init: bool = False,
                 err_exit: bool = False,
                 err_close: bool = False
                 ) -> None:
        self.err_init = err_init
        self.err_exit = err_exit
        self.err_close = err_close

    async def component_init(self):
        if self.err_init:
            raise ServerError("test")

    async def on_exit(self):
        if self.err_exit:
            raise ServerError("test")

    async def close(self):
        if self.err_close:
            raise ServerError("test")

class MockWebsocket:
    def __init__(self, fut: asyncio.Future) -> None:
        self.future = fut

    def queue_message(self, data: str):
        self.future.set_result(data)
